from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Any, ClassVar, NoReturn, Sequence, TypeAlias

from vsexprtools import ExprOp, ExprToken, norm_expr
from vstools import (
    ColorRange, CustomRuntimeError, CustomValueError, FuncExceptT, PlanesT, T, check_variable, core, get_lowest_values,
    get_peak_value, get_peak_values, get_subclasses, inject_self, join, normalize_planes, plane, vs
)

from ..exceptions import UnknownEdgeDetectError, UnknownRidgeDetectError

__all__ = [
    'EdgeDetect', 'EdgeDetectT',
    'RidgeDetect', 'RidgeDetectT',

    'MatrixEdgeDetect', 'SingleMatrix', 'EuclidianDistance',

    'Max',

    'get_all_edge_detects',
    'get_all_ridge_detect',
]


class _Feature(Enum):
    EDGE = auto()
    RIDGE = auto()


class BaseDetect:
    @staticmethod
    def from_param(
        cls: type[T],
        value: str | type[T] | T | None,
        exception_cls: type[CustomValueError],
        excluded: Sequence[type[T]] = [],
        func_except: FuncExceptT | None = None
    ) -> type[T]:
        if isinstance(value, str):
            all_edge_detects = get_subclasses(EdgeDetect, excluded)
            search_str = value.lower().strip()

            for edge_detect_cls in all_edge_detects:
                if edge_detect_cls.__name__.lower() == search_str:
                    return edge_detect_cls  # type: ignore

            raise exception_cls(func_except or cls.from_param, value)  # type: ignore

        if issubclass(value, cls):  # type: ignore
            return value  # type: ignore

        if isinstance(value, cls):
            return value.__class__

        return cls

    @staticmethod
    def ensure_obj(
        cls: type[T],
        value: str | type[T] | T | None,
        exception_cls: type[CustomValueError],
        excluded: Sequence[type[T]] = [],
        func_except: FuncExceptT | None = None
    ) -> T:
        new_edge_detect: T | None = None

        if not isinstance(value, cls):
            try:
                new_edge_detect = cls.from_param(value, func_except)()  # type: ignore
            except Exception:
                ...
        else:
            new_edge_detect = value

        if new_edge_detect is None:
            new_edge_detect = cls()

        if new_edge_detect.__class__ in excluded:
            raise exception_cls(
                func_except or cls.ensure_obj, new_edge_detect.__class__,  # type: ignore
                'This {cls_name} can\'t be instantiated to be used!',
                cls_name=new_edge_detect.__class__
            )

        return new_edge_detect


class EdgeDetect(ABC):
    """Abstract edge detection interface."""

    _bits: int

    @classmethod
    def from_param(
        cls: type[EdgeDetect], edge_detect: EdgeDetectT | None = None, func_except: FuncExceptT | None = None
    ) -> type[EdgeDetect]:
        return BaseDetect.from_param(cls, edge_detect, UnknownEdgeDetectError, [], func_except)

    @classmethod
    def ensure_obj(
        cls: type[EdgeDetect], edge_detect: EdgeDetectT | None = None, func_except: FuncExceptT | None = None
    ) -> EdgeDetect:
        return BaseDetect.ensure_obj(cls, edge_detect, UnknownEdgeDetectError, [], func_except)

    @inject_self
    def edgemask(
        self, clip: vs.VideoNode, lthr: float = 0.0, hthr: float | None = None, multi: float = 1.0,
        clamp: bool | tuple[float, float] | list[tuple[float, float]] = False, planes: PlanesT = None,
        **kwargs: Any
    ) -> vs.VideoNode:
        """
        Makes edge mask based on convolution kernel.
        The resulting mask can be thresholded with lthr, hthr and multiplied with multi.

        :param clip:            Source clip
        :param lthr:            Low threshold. Anything below lthr will be set to 0
        :param hthr:            High threshold. Anything above hthr will be set to the range max
        :param multi:           Multiply all pixels by this before thresholding
        :param clamp:           Clamp to TV or full range if True or specified range `(low, high)`

        :return:                Mask clip
        """
        return self._mask(clip, lthr, hthr, multi, clamp, _Feature.EDGE, planes, **kwargs)

    def _mask(
        self,
        clip: vs.VideoNode,
        lthr: float = 0.0, hthr: float | None = None,
        multi: float = 1.0,
        clamp: bool | tuple[float, float] | list[tuple[float, float]] = False,
        feature: _Feature = _Feature.EDGE, planes: PlanesT = None, **kwargs: Any
    ) -> vs.VideoNode:
        assert check_variable(clip, self.__class__)

        self._bits = clip.format.bits_per_sample
        peak = get_peak_value(clip)
        hthr = peak if hthr is None else hthr

        planes = normalize_planes(clip, planes)

        wclip = plane(clip, planes[0]) if len(planes) == 1 else clip

        clip_p = self._preprocess(wclip)

        try:
            if feature == _Feature.EDGE:
                mask = self._compute_edge_mask(clip_p, **kwargs)
            elif feature == _Feature.RIDGE:
                if not isinstance(self, RidgeDetect):
                    raise RuntimeError
                mask = self._compute_ridge_mask(clip_p, **kwargs)
        except Exception as e:
            raise CustomRuntimeError('There was an error processing the mask! Are you using an abstract class?') from e

        mask = self._postprocess(mask)

        if multi != 1:
            mask = ExprOp.MUL(mask, suffix=multi, planes=planes)

        if lthr > 0 or hthr < peak:
            mask = norm_expr(mask, f'x {hthr} > {ExprToken.RangeMax} x {lthr} < 0 x ? ?', planes)

        if clamp:
            if clamp is True:
                crange = ColorRange.from_video(clip)
                clamp = list(zip(get_lowest_values(mask, crange), get_peak_values(mask, crange)))

            if isinstance(clamp, list):
                mask = norm_expr(mask, [ExprOp.clamp(*c, c='x') for c in clamp], planes)
            elif isinstance(clamp, tuple):
                mask = ExprOp.clamp(*clamp, c='x')(mask, planes=planes)

        assert mask.format

        if mask.format.num_planes != clip.format.num_planes:
            return join({None: clip.std.BlankClip(color=[0] * clip.format.num_planes, keep=True), planes[0]: mask})

        return mask

    @abstractmethod
    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        raise NotImplementedError

    def _preprocess(self, clip: vs.VideoNode) -> vs.VideoNode:
        return clip

    def _postprocess(self, clip: vs.VideoNode) -> vs.VideoNode:
        return clip


class MatrixEdgeDetect(EdgeDetect):
    matrices: ClassVar[Sequence[Sequence[float]]]
    divisors: ClassVar[Sequence[float] | None] = None
    mode_types: ClassVar[Sequence[str] | None] = None

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return self._merge_edge([
            clip.std.Convolution(matrix=mat, divisor=div, saturate=False, mode=mode)
            for mat, div, mode in zip(self._get_matrices(), self._get_divisors(), self._get_mode_types())
        ])

    def _compute_ridge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        def _x(c: vs.VideoNode) -> vs.VideoNode:
            return c.std.Convolution(matrix=self._get_matrices()[0], divisor=self._get_divisors()[0])

        def _y(c: vs.VideoNode) -> vs.VideoNode:
            return c.std.Convolution(matrix=self._get_matrices()[1], divisor=self._get_divisors()[1])

        x = _x(clip)
        y = _y(clip)
        xx = _x(x)
        yy = _y(y)
        xy = _x(x)
        return self._merge_ridge([xx, yy, xy])

    @abstractmethod
    def _merge_edge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode:
        raise NotImplementedError

    @abstractmethod
    def _merge_ridge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode | NoReturn:
        raise NotImplementedError

    def _get_matrices(self) -> Sequence[Sequence[float]]:
        return self.matrices

    def _get_divisors(self) -> Sequence[float]:
        return self.divisors if self.divisors else [0.0] * len(self._get_matrices())

    def _get_mode_types(self) -> Sequence[str]:
        return self.mode_types if self.mode_types else ['s'] * len(self._get_matrices())

    def _postprocess(self, clip: vs.VideoNode) -> vs.VideoNode:
        if len(self.matrices[0]) > 9 or (self.mode_types and self.mode_types[0] != 's'):
            clip = clip.std.Crop(
                right=clip.format.subsampling_w * 2 if clip.format and clip.format.subsampling_w != 0 else 2
            ).resize.Point(clip.width, src_width=clip.width)
        return clip


class RidgeDetect(MatrixEdgeDetect):
    @classmethod
    def from_param(  # type: ignore
        cls: type[RidgeDetect], edge_detect: RidgeDetectT | None = None, func_except: FuncExceptT | None = None
    ) -> type[RidgeDetect]:
        return BaseDetect.from_param(cls, edge_detect, UnknownRidgeDetectError, [], func_except)

    @classmethod
    def ensure_obj(  # type: ignore
        cls: type[RidgeDetect], edge_detect: RidgeDetectT | None = None, func_except: FuncExceptT | None = None
    ) -> RidgeDetect:
        return BaseDetect.ensure_obj(cls, edge_detect, UnknownRidgeDetectError, [], func_except)

    @inject_self
    def ridgemask(
        self, clip: vs.VideoNode, lthr: float = 0.0, hthr: float | None = None, multi: float = 1.0,
        clamp: bool | tuple[float, float] | list[tuple[float, float]] = False, **kwargs: Any
    ) -> vs.VideoNode | NoReturn:
        """
        Makes ridge mask based on convolution kernel.
        The resulting mask can be thresholded with lthr, hthr and multiplied with multi.

        :param clip:            Source clip
        :param lthr:            Low threshold. Anything below lthr will be set to 0
        :param hthr:            High threshold. Anything above hthr will be set to the range max
        :param multi:           Multiply all pixels by this before thresholding
        :param clamp:           Clamp to TV or full range if True or specified range `(low, high)`

        :return:                Mask clip
        """
        return self._mask(clip, lthr, hthr, multi, clamp, _Feature.RIDGE, **kwargs)

    def _merge_ridge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode:
        return core.std.Expr(clips, 'x y * 2 * -1 * x dup * z dup * 4 * + y dup * + + sqrt x y + +')


class SingleMatrix(MatrixEdgeDetect, ABC):
    def _merge_edge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode:
        return clips[0]

    def _merge_ridge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode | NoReturn:
        raise NotImplementedError


class EuclidianDistance(MatrixEdgeDetect, ABC):
    def _merge_edge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode:
        return core.std.Expr(clips, 'x x * y y * + sqrt')

    def _merge_ridge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode | NoReturn:
        raise NotImplementedError


class Max(MatrixEdgeDetect, ABC):
    def _merge_edge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode:
        return ExprOp.MAX.combine(*clips)

    def _merge_ridge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode | NoReturn:
        raise NotImplementedError


EdgeDetectT: TypeAlias = type[EdgeDetect] | EdgeDetect | str  # type: ignore
RidgeDetectT: TypeAlias = type[RidgeDetect] | RidgeDetect | str  # type: ignore


def get_all_edge_detects(
    clip: vs.VideoNode,
    lthr: float = 0.0, hthr: float | None = None,
    multi: float = 1.0,
    clamp: bool | tuple[float, float] | list[tuple[float, float]] = False
) -> list[vs.VideoNode]:
    """
    Returns all the EdgeDetect subclasses

    :param clip:        Source clip
    :param lthr:        See :py:func:`EdgeDetect.get_mask`
    :param hthr:        See :py:func:`EdgeDetect.get_mask`
    :param multi:       See :py:func:`EdgeDetect.get_mask`
    :param clamp:       Clamp to TV or full range if True or specified range `(low, high)`

    :return:            A list edge masks
    """
    def _all_subclasses(cls: type[EdgeDetect] = EdgeDetect) -> set[type[EdgeDetect]]:
        return set(cls.__subclasses__()).union(s for c in cls.__subclasses__() for s in _all_subclasses(c))

    all_subclasses = {
        s for s in _all_subclasses()
        if s.__name__ not in {
            'MatrixEdgeDetect', 'RidgeDetect', 'SingleMatrix', 'EuclidianDistance', 'Max',
            'Matrix1D', 'SavitzkyGolay', 'SavitzkyGolayNormalise',
            'Matrix2x2', 'Matrix3x3', 'Matrix5x5'
        }
    }
    return [
        edge_detect().edgemask(clip, lthr, hthr, multi, clamp).text.Text(edge_detect.__name__)
        for edge_detect in sorted(all_subclasses, key=lambda x: x.__name__)
    ]


def get_all_ridge_detect(
    clip: vs.VideoNode, lthr: float = 0.0, hthr: float | None = None, multi: float = 1.0,
    clamp: bool | tuple[float, float] | list[tuple[float, float]] = False
) -> list[vs.VideoNode]:
    """
    Returns all the RidgeDetect subclasses

    :param clip:        Source clip
    :param lthr:        See :py:func:`EdgeDetect.get_mask`
    :param hthr:        See :py:func:`EdgeDetect.get_mask`
    :param multi:       See :py:func:`EdgeDetect.get_mask`
    :param clamp:       Clamp to TV or full range if True or specified range `(low, high)`

    :return:            A list edge masks
    """
    def _all_subclasses(cls: type[RidgeDetect] = RidgeDetect) -> set[type[RidgeDetect]]:
        return set(cls.__subclasses__()).union(s for c in cls.__subclasses__() for s in _all_subclasses(c))

    all_subclasses = {
        s for s in _all_subclasses()
        if s.__name__ not in {
            'MatrixEdgeDetect', 'RidgeDetect', 'SingleMatrix', 'EuclidianDistance', 'Max',
            'Matrix1D', 'SavitzkyGolay', 'SavitzkyGolayNormalise',
            'Matrix2x2', 'Matrix3x3', 'Matrix5x5'
        }
    }
    return [
        edge_detect().ridgemask(clip, lthr, hthr, multi, clamp).text.Text(edge_detect.__name__)
        for edge_detect in sorted(all_subclasses, key=lambda x: x.__name__)
    ]



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any

from vsexprtools import ExprOp, ExprToken, norm_expr
from vskernels import Catrom
from vsrgtools.util import mean_matrix
from vstools import (
    CustomOverflowError, FileNotExistsError, VSFunction, check_variable, core, depth, fallback, get_neutral_value,
    get_neutral_values, get_y, insert_clip, iterate, normalize_ranges, replace_ranges, scale_8bit, scale_value, split,
    vs, vs_object
)

from .abstract import DeferredMask, GeneralMask
from .edge import Sobel
from .morpho import Morpho
from .types import GenericMaskT
from .utils import normalize_mask

__all__ = [
    'HardsubManual',

    'HardsubMask',
    'HardsubSignFades',
    'HardsubSign',
    'HardsubLine',
    'HardsubLineFade',
    'HardsubASS',

    'bounded_dehardsub',
    'diff_hardsub_mask',

    'get_all_sign_masks'
]


@dataclass
class HardsubManual(GeneralMask, vs_object):
    path: str | Path
    processing: VSFunction = core.lazy.std.Binarize  # type: ignore

    def __post_init__(self) -> None:
        if not (path := Path(self.path)).is_dir():
            raise FileNotExistsError('"path" must be an existing path directory!', self.get_mask)

        files = [file.stem for file in path.glob('*')]

        self.clips = [
            core.imwri.Read(file) for file in files
        ]

        self.ranges = [
            (other[-1] if other else end, end)
            for (*other, end) in (map(int, name.split('_')) for name in files)
        ]

    def get_mask(self, clip: vs.VideoNode, *args: Any) -> vs.VideoNode:  # type: ignore[override]
        assert check_variable(clip, self.get_mask)

        mask = clip.std.BlankClip(
            format=clip.format.replace(color_family=vs.GRAY, subsampling_h=0, subsampling_w=0).id,
            keep=True, color=0
        )

        for maskclip, (start_frame, end_frame) in zip(self.clips, self.ranges):
            maskclip = maskclip.std.AssumeFPS(clip).resize.Point(format=mask.format.id)  # type: ignore
            maskclip = self.processing(maskclip).std.Loop(end_frame - start_frame + 1)

            mask = insert_clip(mask, maskclip, start_frame)

        return mask

    def __vs_del__(self, core_id: int) -> None:
        super().__vs_del__(core_id)

        self.clips.clear()


class HardsubMask(DeferredMask):
    bin_thr: float = 0.75

    def get_progressive_dehardsub(
        self, hardsub: vs.VideoNode, ref: vs.VideoNode, partials: list[vs.VideoNode]
    ) -> tuple[list[vs.VideoNode], list[vs.VideoNode]]:
        """
        Dehardsub using multiple superior hardsubbed sources and one inferior non-subbed source.

        :param hardsub:  Hardsub master source (eg Wakanim RU dub).
        :param ref:      Non-subbed reference source (eg CR, Funi, Amazon).
        :param partials: Sources to use for partial dehardsubbing (eg Waka DE, FR, SC).

        :return:         Dehardsub stages and masks used for progressive dehardsub.
        """

        masks = [self.get_mask(hardsub, ref)]
        partials_dehardsubbed = [hardsub]
        dehardsub_masks = []
        partials = partials + [ref]

        assert masks[-1].format is not None

        thresh = scale_value(self.bin_thr, 32, masks[-1])

        for p in partials:
            masks.append(
                ExprOp.SUB.combine(masks[-1], self.get_mask(p, ref))
            )
            dehardsub_masks.append(
                iterate(core.akarin.Expr([masks[-1]], f"x {thresh} < 0 x ?"), core.std.Maximum, 4).std.Inflate()
            )
            partials_dehardsubbed.append(
                partials_dehardsubbed[-1].std.MaskedMerge(p, dehardsub_masks[-1])
            )

            masks[-1] = masks[-1].std.MaskedMerge(masks[-1].std.Invert(), masks[-2])

        return partials_dehardsubbed, dehardsub_masks

    def apply_dehardsub(
        self, hardsub: vs.VideoNode, ref: vs.VideoNode, partials: list[vs.VideoNode] | None = None
    ) -> vs.VideoNode:
        if partials:
            partials_dehardsubbed, _ = self.get_progressive_dehardsub(hardsub, ref, partials)
            dehardsub = partials_dehardsubbed[-1]
        else:
            dehardsub = hardsub.std.MaskedMerge(ref, self.get_mask(hardsub, ref))

        return replace_ranges(hardsub, dehardsub, self.ranges)


class HardsubSignFades(HardsubMask):
    highpass: float
    expand: int
    edgemask: GenericMaskT

    def __init__(
        self, *args: Any, highpass: float = 0.0763, expand: int = 8, edgemask: GenericMaskT = Sobel, **kwargs: Any
    ) -> None:
        self.highpass = highpass
        self.expand = expand
        self.edgemask = edgemask

        super().__init__(*args, **kwargs)

    def _mask(self, clip: vs.VideoNode, ref: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        clipedge, refedge = (
            normalize_mask(self.edgemask, x, **kwargs).std.Convolution(mean_matrix)
            for x in (clip, ref)
        )

        highpass = scale_value(self.highpass, 32, clip)

        mask = norm_expr(
            [clipedge, refedge], f'x y - {highpass} < 0 {ExprToken.RangeMax} ?'
        ).std.Median()

        return Morpho.inflate(Morpho.maximum(mask, iterations=self.expand), iterations=4)


class HardsubSign(HardsubMask):
    """
    Hardsub scenefiltering helper using `Zastin <https://github.com/kgrabs>`_'s hardsub mask.

    :param thresh:  Binarization threshold, [0, 1] (Default: 0.06).
    :param expand:  std.Maximum iterations (Default: 8).
    :param inflate: std.Inflate iterations (Default: 7).
    """

    thresh: float
    minimum: int
    expand: int
    inflate: int

    def __init__(
        self, *args: Any, thresh: float = 0.06, minimum: int = 1, expand: int = 8, inflate: int = 7, **kwargs: Any
    ) -> None:
        self.thresh = thresh
        self.minimum = minimum
        self.expand = expand
        self.inflate = inflate
        super().__init__(*args, **kwargs)

    def _mask(self, clip: vs.VideoNode, ref: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        hsmf = core.akarin.Expr([clip, ref], 'x y - abs')
        hsmf = hsmf.resize.Point(format=clip.format.replace(subsampling_w=0, subsampling_h=0).id)  # type: ignore

        hsmf = core.akarin.Expr(split(hsmf), "x y z max max")

        hsmf = Morpho.binarize(hsmf, self.thresh)
        hsmf = Morpho.minimum(hsmf, iterations=self.minimum)
        hsmf = Morpho.maximum(hsmf, iterations=self.expand)
        hsmf = Morpho.inflate(hsmf, iterations=self.inflate)

        return hsmf.std.Limiter()


class HardsubLine(HardsubMask):
    expand: int | None

    def __init__(self, *args: Any, expand: int | None = None, **kwargs: Any) -> None:
        self.expand = expand

        super().__init__(*args, **kwargs)

    def _mask(self, clip: vs.VideoNode, ref: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        assert clip.format

        expand_n = fallback(self.expand, clip.width // 200)

        y_range = scale_8bit(clip, 219) if clip.format.sample_type == vs.INTEGER else 1
        uv_range = scale_8bit(clip, 224) if clip.format.sample_type == vs.INTEGER else 1
        offset = scale_8bit(clip, 16) if clip.format.sample_type == vs.INTEGER else 0

        uv_abs = ' abs ' if clip.format.sample_type == vs.FLOAT else f' {get_neutral_value(clip)} - abs '
        yexpr = f'x y - abs {y_range * 0.7} > 255 0 ?'
        uv_thr = uv_range * 0.8
        uvexpr = f'x {uv_abs} {uv_thr} < y {uv_abs} {uv_thr} < and 255 0 ?'

        upper = y_range * 0.8 + offset
        lower = y_range * 0.2 + offset
        mindiff = y_range * 0.1

        difexpr = f'x {upper} > x {lower} < or x y - abs {mindiff} > and 255 0 ?'

        right = core.resize.Point(clip, src_left=4)

        subedge = norm_expr(
            [clip, right], (yexpr, uvexpr), format=clip.format.replace(sample_type=vs.INTEGER, bits_per_sample=8)
        )

        subedge = ExprOp.MIN(Catrom.resample(subedge, vs.YUV444P8), split_planes=True)

        clip_y, ref_y = get_y(clip), depth(get_y(ref), clip)

        clips = [clip_y.std.Convolution(mean_matrix), ref_y.std.Convolution(mean_matrix)]
        diff = core.std.Expr(clips, difexpr, vs.GRAY8).std.Maximum().std.Maximum()

        mask: vs.VideoNode = core.misc.Hysteresis(subedge, diff)
        mask = iterate(mask, core.std.Maximum, expand_n)
        mask = mask.std.Inflate().std.Inflate().std.Convolution(mean_matrix)

        return depth(mask, clip, range_in=1, range_out=1)


class HardsubLineFade(HardsubLine):
    def __init__(self, *args: Any, refframe: float = 0.5, **kwargs: Any) -> None:
        if refframe < 0 or refframe > 1:
            raise CustomOverflowError('"refframe" must be between 0 and 1!', self.__class__)

        self.ref_float = refframe

        super().__init__(*args, refframes=None, **kwargs)

    def get_mask(self, clip: vs.VideoNode, ref: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:  # type: ignore
        self.refframes = [
            r[0] + round((r[1] - r[0]) * self.ref_float)
            for r in normalize_ranges(ref, self.ranges)
        ]

        return super().get_mask(clip, ref)


class HardsubASS(HardsubMask):
    filename: str
    fontdir: str | None
    shift: int | None

    def __init__(
        self, filename: str, *args: Any, fontdir: str | None = None, shift: int | None = None, **kwargs: Any
    ) -> None:
        self.filename = filename
        self.fontdir = fontdir
        self.shift = shift
        super().__init__(*args, **kwargs)

    def _mask(self, clip: vs.VideoNode, ref: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        ref = ref[0] * self.shift + ref if self.shift else ref
        mask: vs.VideoNode = ref.sub.TextFile(  # type: ignore[attr-defined]
            self.filename, fontdir=self.fontdir, blend=False
        )[1]
        mask = mask[self.shift:] if self.shift else mask
        mask = mask.std.Binarize(1)
        mask = iterate(mask, core.std.Maximum, 3)
        mask = iterate(mask, core.std.Inflate, 3)
        return mask.std.Limiter()


def bounded_dehardsub(
    hrdsb: vs.VideoNode, ref: vs.VideoNode, signs: list[HardsubMask], partials: list[vs.VideoNode] | None = None
) -> vs.VideoNode:
    for sign in signs:
        hrdsb = sign.apply_dehardsub(hrdsb, ref, partials)

    return hrdsb


def diff_hardsub_mask(a: vs.VideoNode, b: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
    assert check_variable(a, diff_hardsub_mask)
    assert check_variable(b, diff_hardsub_mask)

    return a.std.BlankClip(color=get_neutral_values(a), keep=True).std.MaskedMerge(
        a.std.MakeDiff(b), HardsubLine(**kwargs).get_mask(a, b)
    )


def get_all_sign_masks(hrdsb: vs.VideoNode, ref: vs.VideoNode, signs: list[HardsubMask]) -> vs.VideoNode:
    assert check_variable(hrdsb, get_all_sign_masks)
    assert check_variable(ref, get_all_sign_masks)

    mask = ref.std.BlankClip(
        format=ref.format.replace(color_family=vs.GRAY, subsampling_w=0, subsampling_h=0).id, keep=True
    )

    for sign in signs:
        mask = replace_ranges(mask, ExprOp.ADD.combine(mask, sign.get_mask(hrdsb, ref)), sign.ranges)

    return mask.std.Limiter()



"""2D matrices 3x3"""

from __future__ import annotations

from abc import ABC

from ._abstract import EdgeDetect, EuclidianDistance, RidgeDetect

__all__ = [
    'Matrix2x2',
    'Roberts'
]


class Matrix2x2(EdgeDetect, ABC):
    ...


class Roberts(RidgeDetect, EuclidianDistance, Matrix2x2):
    """Lawrence Roberts operator. 2x2 matrices computed in 3x3 matrices."""

    matrices = [
        [0, 0, 0, 0, 1, 0, 0, 0, -1],
        [0, 0, 0, 0, 0, 1, 0, -1, 0]
    ]



from __future__ import annotations

from typing import Any

from vstools import CustomValueError, FuncExceptT

__all__ = [
    'UnknownEdgeDetectError',
    'UnknownRidgeDetectError',
]


class UnknownEdgeDetectError(CustomValueError):
    """Raised when an unknown edge detect is passed."""

    def __init__(
        self, func: FuncExceptT, edge_detect: str, message: str = 'Unknown concrete edge detector "{edge_detect}"!',
        **kwargs: Any
    ) -> None:
        super().__init__(message, func, edge_detect=edge_detect, **kwargs)


class UnknownRidgeDetectError(CustomValueError):
    """Raised when an unknown ridge detect is passed."""

    def __init__(
        self, func: FuncExceptT, ridge_detect: str, message: str = 'Unknown concrete ridge detector "{ridge_detect}"!',
        **kwargs: Any
    ) -> None:
        super().__init__(message, func, ridge_detect=ridge_detect, **kwargs)



from __future__ import annotations

from typing import Any, Iterable

from vsexprtools import ExprOp, aka_expr_available, norm_expr
from vskernels import Bilinear, Kernel, KernelT
from vsrgtools import box_blur, gauss_blur
from vstools import (
    CustomValueError, FrameRangeN, FrameRangesN, FuncExceptT, check_variable, check_variable_format, flatten,
    get_peak_value, insert_clip, replace_ranges, split, vs, depth
)

from .edge import EdgeDetect, RidgeDetect
from .types import GenericMaskT
from .abstract import GeneralMask

__all__ = [
    'max_planes',

    'region_rel_mask', 'region_abs_mask',

    'squaremask', 'replace_squaremask', 'freeze_replace_squaremask',

    'normalize_mask'
]


def max_planes(*_clips: vs.VideoNode | Iterable[vs.VideoNode], resizer: KernelT = Bilinear) -> vs.VideoNode:
    clips = list[vs.VideoNode](flatten(_clips))  # type: ignore

    assert check_variable_format((model := clips[0]), max_planes)

    resizer = Kernel.ensure_obj(resizer, max_planes)

    width, height, fmt = model.width, model.height, model.format.replace(subsampling_w=0, subsampling_h=0)

    return ExprOp.MAX.combine(
        split(resizer.scale(clip, width, height, format=fmt)) for clip in clips
    )


def _get_region_expr(
    clip: vs.VideoNode | vs.VideoFrame, left: int = 0, right: int = 0, top: int = 0, bottom: int = 0,
    replace: str | int = 0, rel: bool = False
) -> str:
    right, bottom = right + 1, bottom + 1

    if isinstance(replace, int):
        replace = f'x {replace}'

    if rel:
        return f'X {left} < X {right} > or Y {top} < Y {bottom} > or or {replace} ?'

    return f'X {left} < X {clip.width - right} > or Y {top} < Y {clip.height - bottom} > or or {replace} ?'


def region_rel_mask(clip: vs.VideoNode, left: int = 0, right: int = 0, top: int = 0, bottom: int = 0) -> vs.VideoNode:
    if aka_expr_available:
        return norm_expr(clip, _get_region_expr(clip, left, right, top, bottom, 0), force_akarin=region_rel_mask)

    return clip.std.Crop(left, right, top, bottom).std.AddBorders(left, right, top, bottom)


def region_abs_mask(clip: vs.VideoNode, width: int, height: int, left: int = 0, top: int = 0) -> vs.VideoNode:
    def _crop(w: int, h: int) -> vs.VideoNode:
        return clip.std.CropAbs(width, height, left, top).std.AddBorders(
            left, w - width - left, top, h - height - top
        )

    if 0 in {clip.width, clip.height}:
        if aka_expr_available:
            return norm_expr(
                clip, _get_region_expr(clip, left, left + width, top, top + height, 0, True),
                force_akarin=region_rel_mask
            )

        return clip.std.FrameEval(lambda f, n: _crop(f.width, f.height), clip)

    return region_rel_mask(clip, left, clip.width - width - left, top, clip.height - height - top)


def squaremask(
    clip: vs.VideoNode, width: int, height: int, offset_x: int, offset_y: int, invert: bool = False,
    func: FuncExceptT | None = None
) -> vs.VideoNode:
    """
    Create a square used for simple masking.

    This is a fast and simple mask that's useful for very rough and simple masking.

    :param clip:        The clip to process.
    :param width:       The width of the square. This must be less than clip.width - offset_x.
    :param height:      The height of the square. This must be less than clip.height - offset_y.
    :param offset_x:    The location of the square, offset from the left side of the frame.
    :param offset_y:    The location of the square, offset from the top of the frame.
    :param invert:      Invert the mask. This means everything *but* the defined square will be masked.
                        Default: False.
    :param func:        Function returned for custom error handling.
                        This should only be set by VS package developers.
                        Default: :py:func:`squaremask`.

    :return:            A mask in the shape of a square.
    """
    func = func or squaremask

    assert check_variable(clip, func)

    mask_format = clip.format.replace(color_family=vs.GRAY, subsampling_w=0, subsampling_h=0)

    if offset_x + width > clip.width or offset_y + height > clip.height:
        raise CustomValueError('mask exceeds clip size!')

    base_clip = clip.std.BlankClip(
        width, height, mask_format.id, 1, color=0 if aka_expr_available else get_peak_value(clip), keep=True
    )

    if aka_expr_available:
        mask = norm_expr(
            base_clip, _get_region_expr(
                clip, offset_x, clip.width - width - offset_x, offset_y, clip.height - height - offset_y,
                'range_max x' if invert else 'x range_max'
            ), force_akarin=func
        )
    else:
        mask = base_clip.std.AddBorders(
            offset_x, clip.width - width - offset_x, offset_y, clip.height - height - offset_y
        )
        if invert:
            mask = mask.std.Invert()

    if clip.num_frames == 1:
        return mask

    return mask.std.Loop(clip.num_frames)


def replace_squaremask(
    clipa: vs.VideoNode, clipb: vs.VideoNode, mask_params: tuple[int, int, int, int],
    ranges: FrameRangeN | FrameRangesN | None = None, blur_sigma: int | float | None = None,
    invert: bool = False, func: FuncExceptT | None = None, show_mask: bool = False
) -> vs.VideoNode:
    """
    Replace an area of the frame with another clip using a simple square mask.

    This is a convenience wrapper merging square masking and framerange replacing functionalities
    into one function, along with additional utilities such as blurring.

    :param clipa:           Base clip to process.
    :param clipb:           Clip to mask on top of `clipa`.
    :param mask_params:     Parameters passed to `squaremask`. Expects a tuple of (width, height, offset_x, offset_y).
    :param ranges:          Frameranges to replace with the masked clip. If `None`, replaces the entire clip.
                            Default: None.
    :param blur_sigma:      Post-blurring of the mask to help hide hard edges.
                            If you pass an int, a :py:func:`box_blur` will be used.
                            Passing a float will use a :py:func:`gauss_blur` instead.
                            Default: None.
    :param invert:          Invert the mask. This means everything *but* the defined square will be masked.
                            Default: False.
    :param func:            Function returned for custom error handling.
                            This should only be set by VS package developers.
                            Default: :py:func:`squaremask`.
    :param show_mask:       Return the mask instead of the masked clip.

    :return:                Clip with a squaremask applied, and optionally set to specific frameranges.
    """
    func = func or replace_squaremask

    assert check_variable(clipa, func) and check_variable(clipb, func)

    mask = squaremask(clipb[0], *mask_params, invert, func)

    if ranges is None:
        ranges = [(None, None)]

    if isinstance(blur_sigma, int):
        mask = box_blur(mask, blur_sigma)
    elif isinstance(blur_sigma, float):
        mask = gauss_blur(mask, blur_sigma)

    mask = mask.std.Loop(clipa.num_frames)

    if show_mask:
        return mask

    merge = clipa.std.MaskedMerge(clipb, mask)

    return replace_ranges(clipa, merge, ranges)


def freeze_replace_squaremask(
    mask: vs.VideoNode, insert: vs.VideoNode, mask_params: tuple[int, int, int, int],
    frame: int, frame_range: tuple[int, int]
) -> vs.VideoNode:
    start, end = frame_range

    masked_insert = replace_squaremask(mask[frame], insert[frame], mask_params)

    return insert_clip(mask, masked_insert * (end - start + 1), start)


def normalize_mask(
    mask: GenericMaskT, clip: vs.VideoNode, ref: vs.VideoNode | None = None,
    *, ridge: bool = False, **kwargs: Any
) -> vs.VideoNode:
    if isinstance(mask, str):
        mask = EdgeDetect.ensure_obj(mask)

    if isinstance(mask, type):
        mask = mask()

    if isinstance(mask, RidgeDetect) and ridge:
        mask = mask.ridgemask(clip, **kwargs)

    if isinstance(mask, EdgeDetect):
        mask = mask.edgemask(clip, **kwargs)

    if isinstance(mask, GeneralMask):
        mask = mask.get_mask(clip, ref)

    if callable(mask):
        if ref is None:
            raise CustomValueError('This mask function requires a ref to be specified!')

        mask = mask(clip, ref)

    return depth(mask, clip)



from __future__ import annotations

from vsexprtools import ExprOp
from vstools import check_variable, join, normalize_seq, split, vs

from .morpho import Morpho
from .types import Coordinates

__all__ = [
    'range_mask'
]


def range_mask(clip: vs.VideoNode, rad: int = 2, radc: int = 0) -> vs.VideoNode:
    assert check_variable(clip, range_mask)

    def _minmax(clip: vs.VideoNode, iters: int, maxx: bool) -> vs.VideoNode:
        func = Morpho.maximum if maxx else Morpho.minimum

        for i in range(1, iters + 1):
            clip = func(clip, coordinates=Coordinates.from_iter(i))

        return clip

    return join([
        ExprOp.SUB.combine(
            _minmax(plane, r, True),
            _minmax(plane, r, False)
        ) for plane, r in zip(
            split(clip), normalize_seq(radc and [rad, radc] or rad, clip.format.num_planes)
        )
    ]).std.Limiter()



"""2D matrices 5x5"""

from __future__ import annotations

import math
from abc import ABC
from dataclasses import dataclass
from typing import Any, NoReturn, Sequence

from vsexprtools import ExprOp
from vstools import ColorRange, depth, get_depth, join, split, vs

from ..morpho import Morpho
from ..types import XxpandMode
from ._abstract import EdgeDetect, EuclidianDistance, MatrixEdgeDetect, Max, RidgeDetect, SingleMatrix

__all__ = [
    'Matrix3x3',
    # Single matrix
    'Laplacian1', 'Laplacian2', 'Laplacian3', 'Laplacian4',
    'Kayyali',
    # Euclidian Distance
    'Tritical', 'TriticalTCanny',
    'Cross',
    'Prewitt', 'PrewittStd', 'PrewittTCanny',
    'Sobel', 'SobelStd', 'SobelTCanny', 'ASobel',
    'Scharr', 'RScharr', 'ScharrTCanny',
    'Kroon', 'KroonTCanny',
    'FreyChenG41', 'FreyChen',
    # Max
    'Robinson3', 'Robinson5', 'TheToof',
    'Kirsch', 'KirschTCanny',
    # Misc
    'MinMax'
]


class Matrix3x3(EdgeDetect, ABC):
    ...


# Single matrix
class Laplacian1(SingleMatrix, Matrix3x3):
    """Pierre-Simon de Laplace operator 1st implementation."""

    matrices = [[0, -1, 0, -1, 4, -1, 0, -1, 0]]


class Laplacian2(SingleMatrix, Matrix3x3):
    """Pierre-Simon de Laplace operator 2nd implementation."""

    matrices = [[1, -2, 1, -2, 4, -2, 1, -2, 1]]


class Laplacian3(SingleMatrix, Matrix3x3):
    """Pierre-Simon de Laplace operator 3rd implementation."""

    matrices = [[2, -1, 2, -1, -4, -1, 2, -1, 2]]


class Laplacian4(SingleMatrix, Matrix3x3):
    """Pierre-Simon de Laplace operator 4th implementation."""

    matrices = [[-1, -1, -1, -1, 8, -1, -1, -1, -1]]


class Kayyali(SingleMatrix, Matrix3x3):
    """Kayyali operator."""

    matrices = [[6, 0, -6, 0, 0, 0, -6, 0, 6]]


# Euclidian Distance
class Tritical(RidgeDetect, EuclidianDistance, Matrix3x3):
    """
    Operator used in Tritical's original TCanny filter.
    Plain and simple orthogonal first order derivative.
    """

    matrices = [
        [0, 0, 0, -1, 0, 1, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, -1, 0]
    ]


class TriticalTCanny(Matrix3x3, EdgeDetect):
    """
    Operator used in Tritical's original TCanny filter.
    Plain and simple orthogonal first order derivative.
    """

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.tcanny.TCanny(kwargs.pop('sigma', 0), mode=1, op=0, **kwargs)


class Cross(RidgeDetect, EuclidianDistance, Matrix3x3):
    """
    "HotDoG" Operator from AVS ExTools by Dogway.
    Plain and simple cross first order derivative.
    """

    matrices = [
        [1, 0, 0, 0, 0, 0, 0, 0, -1],
        [0, 0, -1, 0, 0, 0, 1, 0, 0]
    ]


class Prewitt(RidgeDetect, EuclidianDistance, Matrix3x3):
    """Judith M. S. Prewitt operator."""

    matrices = [
        [1, 0, -1, 1, 0, -1, 1, 0, -1],
        [1, 1, 1, 0, 0, 0, -1, -1, -1]
    ]


class PrewittStd(Matrix3x3, EdgeDetect):
    """Judith M. S. Prewitt Vapoursynth plugin operator."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.std.Prewitt(**kwargs)


class PrewittTCanny(Matrix3x3, EdgeDetect):
    """Judith M. S. Prewitt TCanny plugin operator."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.tcanny.TCanny(kwargs.pop('sigma', 0), mode=1, op=1, scale=2, **kwargs)


class Sobel(RidgeDetect, EuclidianDistance, Matrix3x3):
    """Sobel–Feldman operator."""

    matrices = [
        [1, 0, -1, 2, 0, -2, 1, 0, -1],
        [1, 2, 1, 0, 0, 0, -1, -2, -1]
    ]


class SobelStd(Matrix3x3, EdgeDetect):
    """Sobel–Feldman Vapoursynth plugin operator."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.std.Sobel(**kwargs)


class SobelTCanny(Matrix3x3, EdgeDetect):
    """Sobel–Feldman Vapoursynth plugin operator."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.tcanny.TCanny(kwargs.pop('sigma', 0), mode=1, op=2, scale=2, **kwargs)


class ASobel(Matrix3x3, EdgeDetect):
    """Modified Sobel–Feldman operator from AWarpSharp."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return (vs.core.warp.ASobel if get_depth(clip) < 32 else vs.core.warpsf.ASobel)(  # type: ignore
            clip, 255, **kwargs
        )


class Scharr(RidgeDetect, EuclidianDistance, Matrix3x3):
    """
    Original H. Scharr optimised operator which attempts
    to achieve the perfect rotational symmetry with coefficients 3 and 10.
    """

    matrices = [
        [-3, 0, 3, -10, 0, 10, -3, 0, 3],
        [-3, -10, -3, 0, 0, 0, 3, 10, 3]
    ]
    divisors = [3, 3]


class RScharr(RidgeDetect, EuclidianDistance, Matrix3x3):
    """
    Refined H. Scharr operator to more accurately calculate
    1st derivatives for a 3x3 kernel with coeffs 47 and 162.
    """

    matrices = [
        [-47, 0, 47, -162, 0, 162, -47, 0, 47],
        [-47, -162, -47, 0, 0, 0, 47, 162, 47]
    ]
    divisors = [47, 47]


class ScharrTCanny(Matrix3x3, EdgeDetect):
    """H. Scharr optimised TCanny Vapoursynth plugin operator."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.tcanny.TCanny(kwargs.pop('sigma', 0), mode=1, op=2, scale=4 / 3, **kwargs)


class Kroon(RidgeDetect, EuclidianDistance, Matrix3x3):
    """Dirk-Jan Kroon operator."""

    matrices = [
        [-17, 0, 17, -61, 0, 61, -17, 0, 17],
        [-17, -61, -17, 0, 0, 0, 17, 61, 17]
    ]
    divisors = [17, 17]


class KroonTCanny(Matrix3x3, EdgeDetect):
    """Dirk-Jan Kroon TCanny Vapoursynth plugin operator."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.tcanny.TCanny(kwargs.pop('sigma', 0), mode=1, op=4, scale=1 / 17, **kwargs)


class FreyChen(MatrixEdgeDetect):
    """Chen Frei operator. 3x3 matrices properly implemented."""

    sqrt2 = math.sqrt(2)
    matrices = [
        [1, sqrt2, 1, 0, 0, 0, -1, -sqrt2, -1],
        [1, 0, -1, sqrt2, 0, -sqrt2, 1, 0, -1],
        [0, -1, sqrt2, 1, 0, -1, -sqrt2, 1, 0],
        [sqrt2, -1, 0, -1, 0, 1, 0, 1, -sqrt2],
        [0, 1, 0, -1, 0, -1, 0, 1, 0],
        [-1, 0, 1, 0, 0, 0, 1, 0, -1],
        [1, -2, 1, -2, 4, -2, 1, -2, 1],
        [-2, 1, -2, 1, 4, 1, -2, 1, -2],
        [1, 1, 1, 1, 1, 1, 1, 1, 1]
    ]
    divisors = [
        2 * sqrt2,
        2 * sqrt2,
        2 * sqrt2,
        2 * sqrt2,
        2,
        2,
        6,
        6,
        3
    ]

    def _preprocess(self, clip: vs.VideoNode) -> vs.VideoNode:
        return depth(clip, 32)

    def _postprocess(self, clip: vs.VideoNode) -> vs.VideoNode:
        return depth(clip, self._bits, range_in=ColorRange.FULL, range_out=ColorRange.FULL)

    def _merge_edge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode:
        M = 'x x * y y * + z z * + a a * +'
        S = f'b b * c c * + d d * + e e * + f f * + {M} +'
        return vs.core.std.Expr(clips, f'{M} {S} / sqrt')

    def _merge_ridge(self, clips: Sequence[vs.VideoNode]) -> vs.VideoNode | NoReturn:
        raise NotImplementedError


class FreyChenG41(RidgeDetect, EuclidianDistance, Matrix3x3):
    """"Chen Frei" operator. 3x3 matrices from G41Fun."""

    matrices = [
        [-7, 0, 7, -10, 0, 10, -7, 0, 7],
        [-7, -10, -7, 0, 0, 0, 7, 10, 7]
    ]
    divisors = [7, 7]


# Max
class Robinson3(Max, Matrix3x3):
    """Robinson compass operator level 3."""

    matrices = [
        [1, 1, 1, 0, 0, 0, -1, -1, -1],
        [1, 1, 0, 1, 0, -1, 0, -1, -1],
        [1, 0, -1, 1, 0, -1, 1, 0, -1],
        [0, -1, -1, 1, 0, -1, 1, 1, 0]
    ]


class Robinson5(Max, Matrix3x3):
    """Robinson compass operator level 5."""

    matrices = [
        [1, 2, 1, 0, 0, 0, -1, -2, -1],
        [2, 1, 0, 1, 0, -1, 0, -1, -2],
        [1, 0, -1, 2, 0, -2, 1, 0, -1],
        [0, -1, -2, 1, 0, -1, 2, 1, 0]
    ]


class TheToof(Max, Matrix3x3):
    """TheToof compass operator from SharpAAMCmod."""

    matrices = [
        [5, 10, 5, 0, 0, 0, -5, -10, -5],
        [10, 5, 0, 5, 0, -5, 0, -5, -10],
        [5, 0, -5, 10, 0, -10, 5, 0, -5],
        [0, -5, -10, 5, 0, -5, 10, 5, 0]
    ]
    divisors = [4] * 4


class Kirsch(Max, Matrix3x3):
    """Russell Kirsch compass operator."""

    matrices = [
        [5, 5, 5, -3, 0, -3, -3, -3, -3],
        [5, 5, -3, 5, 0, -3, -3, -3, -3],
        [5, -3, -3, 5, 0, -3, 5, -3, -3],
        [-3, -3, -3, 5, 0, -3, 5, 5, -3],
        [-3, -3, -3, -3, 0, -3, 5, 5, 5],
        [-3, -3, -3, -3, 0, 5, -3, 5, 5],
        [-3, -3, 5, -3, 0, 5, -3, -3, 5],
        [-3, 5, 5, -3, 0, 5, -3, -3, -3]
    ]


class KirschTCanny(Matrix3x3, EdgeDetect):
    """Russell Kirsch compass TCanny Vapoursynth plugin operator."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.tcanny.TCanny(kwargs.pop('sigma', 0), mode=1, op=5, **kwargs)


# Misc
@dataclass
class MinMax(EdgeDetect):
    """Min/max mask with separate luma/chroma radii."""

    rady: int = 2
    radc: int = 0

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        assert clip.format

        return join([
            ExprOp.SUB.combine(
                Morpho.expand(p, rad, rad, XxpandMode.ELLIPSE, **kwargs),
                Morpho.inpand(p, rad, rad, XxpandMode.ELLIPSE, **kwargs)
            ) if rad > 0 else p for p, rad in zip(
                split(clip), (self.rady, self.radc, self.radc)[:clip.format.num_planes]
            )
        ], clip.format.color_family)



from __future__ import annotations

from typing import Any

from vsexprtools import ExprOp
from vskernels import Bicubic, Catrom, Kernel, KernelT
from vsrgtools import RemoveGrainMode, bilateral, gauss_blur, removegrain
from vstools import (
    CustomValueError, FuncExceptT, KwargsT, check_variable, depth, expect_bits, get_w, get_y, insert_clip, iterate, vs
)

from .edge import ExLaplacian4
from .morpho import Morpho
from .types import XxpandMode

__all__ = [
    'diff_rescale',
    'diff_creditless',
    'diff_creditless_oped',
    'credit_mask',
]


def diff_rescale(
    clip: vs.VideoNode, height: int, kernel: KernelT = Catrom,
    thr: float = 0.216, expand: int = 2, func: FuncExceptT | None = None
) -> vs.VideoNode:
    func = func or diff_rescale

    assert check_variable(clip, func)

    kernel = Kernel.ensure_obj(kernel, func)

    y = get_y(clip)

    pre, bits = expect_bits(y, 32)

    descale = kernel.descale(pre, get_w(height), height)
    rescale = kernel.scale(descale, y.width, y.height)

    diff = ExprOp.mae(y)(pre, rescale)

    mask = iterate(diff, removegrain, 2, RemoveGrainMode.MINMAX_AROUND2)
    mask = mask.std.Expr(f'x 2 4 pow * {thr} < 0 1 ?')

    mask = Morpho.expand(mask, 2 + expand, mode=XxpandMode.ELLIPSE).std.Deflate()

    return depth(mask, bits)


def diff_creditless(
    credit_clip: vs.VideoNode, nc_clip: vs.VideoNode, thr: float,
    start_frame: int = 0, expand: int = 2, *, prefilter: bool | int = False,
    ep_clip: vs.VideoNode | None = None, func: FuncExceptT | None = None, **kwargs: Any
) -> vs.VideoNode:
    func = func or diff_creditless

    assert check_variable(credit_clip, func)

    clips = [credit_clip, nc_clip]

    if prefilter:
        sigma = 5 if prefilter is True else prefilter
        kwargs |= KwargsT(sigmaS=((sigma ** 2 - 1) / 12) ** 0.5, sigmaR=sigma / 10) | kwargs
        clips = [bilateral(c, **kwargs) for c in clips]

    dst_fmt = credit_clip.format.replace(subsampling_w=0, subsampling_h=0)
    diff_fmt = dst_fmt.replace(color_family=vs.GRAY)

    diff = ExprOp.mae(dst_fmt)(
        (Bicubic.resample(c, dst_fmt) for c in clips),
        format=diff_fmt, split_planes=True
    )

    mask = Morpho.binarize(ExLaplacian4.edgemask(diff), thr)
    mask = Morpho.expand(mask, 2 + expand, mode=XxpandMode.ELLIPSE)

    if not ep_clip or ep_clip.num_frames == mask.num_frames:
        return mask

    blank = ep_clip.std.BlankClip(format=diff_fmt.id, keep=True)

    return insert_clip(blank, mask, start_frame)


def diff_creditless_oped(
    ep: vs.VideoNode, ncop: vs.VideoNode, nced: vs.VideoNode,
    opstart: int | None = None, opend: int | None = None,
    edstart: int | None = None, edend: int | None = None,
    func: FuncExceptT | None = None, **kwargs: Any
) -> vs.VideoNode:
    func = func or diff_creditless_oped

    op_mask = ed_mask = None

    kwargs |= KwargsT(thr=25, expand=4, prefilter=False, func=func, ep_clip=ep) | kwargs

    if opstart is not None and opend is not None:
        op_mask = diff_creditless(ep[opstart:opend + 1], ncop[:opend - opstart + 1], opstart, **kwargs)

    if edstart is not None and edend is not None:
        ed_mask = diff_creditless(ep[edstart:edend + 1], nced[:edend - edstart + 1], edstart, **kwargs)

    if op_mask and ed_mask:
        return ExprOp.ADD.combine(op_mask, ed_mask)
    elif op_mask or ed_mask:
        return op_mask or ed_mask  # type: ignore

    raise CustomValueError(
        'You must specify one or both of ("opstart", "opend"), ("edstart", "edend")', func
    )


def credit_mask(
    clip: vs.VideoNode, ref: vs.VideoNode, thr: float,
    blur: float | None = 1.65, prefilter: bool | int = 5,
    expand: int = 8
) -> vs.VideoNode:
    if blur:
        clip = gauss_blur(clip, blur)
        ref = gauss_blur(ref, blur)

    ed_mask = diff_creditless(clip, ref, thr, prefilter=prefilter)

    credit_mask, bits = expect_bits(ed_mask)
    credit_mask = Morpho.erosion(credit_mask, 6)
    credit_mask = iterate(credit_mask, lambda x: x.std.Minimum().std.Maximum(), 8)

    if expand:
        credit_mask = Morpho.dilation(credit_mask, expand)

    credit_mask = Morpho.inflate(credit_mask, 3)

    return depth(credit_mask, bits)



"""1D matrices"""

from __future__ import annotations

from abc import ABC
from typing import Any, Sequence

from vstools import ColorRange, depth, vs

from ._abstract import EdgeDetect, EuclidianDistance

__all__ = [
    'Matrix1D',
    'TEdge', 'TEdgeTedgemask',
    #
    'SavitzkyGolay',
    #
    'SavitzkyGolayDeriv1Quad5',
    'SavitzkyGolayDeriv1Quad7',
    'SavitzkyGolayDeriv1Quad9',
    'SavitzkyGolayDeriv1Quad11',
    'SavitzkyGolayDeriv1Quad13',
    'SavitzkyGolayDeriv1Quad15',
    'SavitzkyGolayDeriv1Quad17',
    'SavitzkyGolayDeriv1Quad19',
    'SavitzkyGolayDeriv1Quad21',
    'SavitzkyGolayDeriv1Quad23',
    'SavitzkyGolayDeriv1Quad25',
    #
    'SavitzkyGolayDeriv1Cubic5',
    'SavitzkyGolayDeriv1Cubic7',
    'SavitzkyGolayDeriv1Cubic9',
    'SavitzkyGolayDeriv1Cubic11',
    'SavitzkyGolayDeriv1Cubic13',
    'SavitzkyGolayDeriv1Cubic15',
    'SavitzkyGolayDeriv1Cubic17',
    'SavitzkyGolayDeriv1Cubic19',
    'SavitzkyGolayDeriv1Cubic21',
    'SavitzkyGolayDeriv1Cubic23',
    'SavitzkyGolayDeriv1Cubic25',
    #
    'SavitzkyGolayDeriv1Quint7',
    'SavitzkyGolayDeriv1Quint9',
    'SavitzkyGolayDeriv1Quint11',
    'SavitzkyGolayDeriv1Quint13',
    'SavitzkyGolayDeriv1Quint15',
    'SavitzkyGolayDeriv1Quint17',
    'SavitzkyGolayDeriv1Quint19',
    'SavitzkyGolayDeriv1Quint21',
    'SavitzkyGolayDeriv1Quint23',
    'SavitzkyGolayDeriv1Quint25',
    #
    'SavitzkyGolayDeriv2Quad5',
    'SavitzkyGolayDeriv2Quad7',
    'SavitzkyGolayDeriv2Quad9',
    'SavitzkyGolayDeriv2Quad11',
    'SavitzkyGolayDeriv2Quad13',
    'SavitzkyGolayDeriv2Quad15',
    'SavitzkyGolayDeriv2Quad17',
    'SavitzkyGolayDeriv2Quad19',
    'SavitzkyGolayDeriv2Quad21',
    'SavitzkyGolayDeriv2Quad23',
    'SavitzkyGolayDeriv2Quad25',
    'SavitzkyGolayDeriv2Quart7',
    'SavitzkyGolayDeriv2Quart9',
    'SavitzkyGolayDeriv2Quart11',
    'SavitzkyGolayDeriv2Quart13',
    'SavitzkyGolayDeriv2Quart15',
    'SavitzkyGolayDeriv2Quart17',
    'SavitzkyGolayDeriv2Quart19',
    'SavitzkyGolayDeriv2Quart21',
    'SavitzkyGolayDeriv2Quart23',
    'SavitzkyGolayDeriv2Quart25',
    #
    'SavitzkyGolayDeriv3Cub5',
    'SavitzkyGolayDeriv3Cub7',
    'SavitzkyGolayDeriv3Cub9',
    'SavitzkyGolayDeriv3Cub11',
    'SavitzkyGolayDeriv3Cub13',
    'SavitzkyGolayDeriv3Cub15',
    'SavitzkyGolayDeriv3Cub17',
    'SavitzkyGolayDeriv3Cub19',
    'SavitzkyGolayDeriv3Cub21',
    'SavitzkyGolayDeriv3Cub23',
    'SavitzkyGolayDeriv3Cub25',
    #
    'SavitzkyGolayDeriv3Quint7',
    'SavitzkyGolayDeriv3Quint9',
    'SavitzkyGolayDeriv3Quint11',
    'SavitzkyGolayDeriv3Quint13',
    'SavitzkyGolayDeriv3Quint15',
    'SavitzkyGolayDeriv3Quint17',
    'SavitzkyGolayDeriv3Quint19',
    'SavitzkyGolayDeriv3Quint21',
    'SavitzkyGolayDeriv3Quint23',
    'SavitzkyGolayDeriv3Quint25',
    #
    'SavitzkyGolayDeriv4Quart7',
    'SavitzkyGolayDeriv4Quart9',
    'SavitzkyGolayDeriv4Quart11',
    'SavitzkyGolayDeriv4Quart13',
    'SavitzkyGolayDeriv4Quart15',
    'SavitzkyGolayDeriv4Quart17',
    'SavitzkyGolayDeriv4Quart19',
    'SavitzkyGolayDeriv4Quart21',
    'SavitzkyGolayDeriv4Quart23',
    'SavitzkyGolayDeriv4Quart25',
    #
    'SavitzkyGolayDeriv5Quint7',
    'SavitzkyGolayDeriv5Quint9',
    'SavitzkyGolayDeriv5Quint11',
    'SavitzkyGolayDeriv5Quint13',
    'SavitzkyGolayDeriv5Quint15',
    'SavitzkyGolayDeriv5Quint17',
    'SavitzkyGolayDeriv5Quint19',
    'SavitzkyGolayDeriv5Quint21',
    'SavitzkyGolayDeriv5Quint23',
    'SavitzkyGolayDeriv5Quint25',
]


class Matrix1D(EdgeDetect, ABC):
    ...


class TEdge(EuclidianDistance, Matrix1D):
    """(TEdgeMasktype=2) Avisynth plugin."""

    matrices = [
        [12, -74, 0, 74, -12],
        [-12, 74, 0, -74, 12]
    ]
    divisors = [62, 62]
    mode_types = ['h', 'v']


class TEdgeTedgemask(Matrix1D, EdgeDetect):
    """(tedgemask.TEdgeMask(threshold=0.0, type=2)) Vapoursynth plugin."""

    def _compute_edge_mask(self, clip: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        return clip.tedgemask.TEdgeMask(threshold=0, type=2)  # type: ignore


class SavitzkyGolay(EuclidianDistance, Matrix1D):
    mode_types = ['h', 'v']


class SavitzkyGolayNormalise(SavitzkyGolay):
    def _preprocess(self, clip: vs.VideoNode) -> vs.VideoNode:
        return depth(clip, 32)

    def _postprocess(self, clip: vs.VideoNode) -> vs.VideoNode:
        return depth(clip, self._bits, range_in=ColorRange.FULL, range_out=ColorRange.FULL)

    def _get_matrices(self) -> Sequence[Sequence[float]]:
        assert self.divisors
        return [[c / div for c in mat] for mat, div in zip(self.matrices, self.divisors)]

    def _get_divisors(self) -> Sequence[float]:
        return [0.0] * len(self._get_matrices())


class SavitzkyGolayDeriv1Quad5(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 5"""

    matrices = [[-2, -1, 0, 1, 2]] * 2
    divisors = [10] * 2


class SavitzkyGolayDeriv1Quad7(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 7"""

    matrices = [[-3, -2, -1, 0, 1, 2, 3]] * 2
    divisors = [28] * 2


class SavitzkyGolayDeriv1Quad9(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 9"""

    matrices = [[-4, -3, -2, -1, 0, 1, 2, 3, 4]] * 2
    divisors = [60] * 2


class SavitzkyGolayDeriv1Quad11(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 11"""

    matrices = [[-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5]] * 2
    divisors = [110] * 2


class SavitzkyGolayDeriv1Quad13(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 13"""

    matrices = [[-6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6]] * 2
    divisors = [182] * 2


class SavitzkyGolayDeriv1Quad15(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 15"""

    matrices = [[-7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7]] * 2
    divisors = [280] * 2


class SavitzkyGolayDeriv1Quad17(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 17"""

    matrices = [[-8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8]] * 2
    divisors = [408] * 2


class SavitzkyGolayDeriv1Quad19(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 19"""

    matrices = [[-9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]] * 2
    divisors = [570] * 2


class SavitzkyGolayDeriv1Quad21(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 21"""

    matrices = [[-10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]] * 2
    divisors = [770] * 2


class SavitzkyGolayDeriv1Quad23(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 23"""

    matrices = [[-11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]] * 2
    divisors = [1012] * 2


class SavitzkyGolayDeriv1Quad25(SavitzkyGolay):
    """Savitzky-Golay first quadratic derivative operator of size 25"""

    matrices = [[-12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]] * 2
    divisors = [1300] * 2


class SavitzkyGolayDeriv1Cubic5(SavitzkyGolay):
    """Savitzky-Golay first cubic/quartic operator of size 5"""

    matrices = [[1, -8, 0, 8, -1]] * 2
    divisors = [12] * 2


class SavitzkyGolayDeriv1Cubic7(SavitzkyGolay):
    """Savitzky-Golay first cubic/quartic derivative operator of size 7"""

    matrices = [[22, -67, -58, 0, 58, 67, -22]] * 2
    divisors = [252] * 2


class SavitzkyGolayDeriv1Cubic9(SavitzkyGolay):
    """Savitzky-Golay first cubic/quartic operator of size 9"""

    matrices = [[86, -142, -193, -126, 0, 126, 193, 142, -86]] * 2
    divisors = [1188] * 2


class SavitzkyGolayDeriv1Cubic11(SavitzkyGolay):
    """Savitzky-Golay first cubic/quartic operator of size 11"""

    matrices = [[300, -294, -532, -503, -296, 0, 296, 503, 532, 294, -300]] * 2
    divisors = [5148] * 2


class SavitzkyGolayDeriv1Cubic13(SavitzkyGolayNormalise):
    """Savitzky-Golay first cubic/quartic operator of size 13"""

    matrices = [[1133, -660, -1578, -1796, -1489, -832, 0, 832, 1489, 1796, 1578, 660, -1133]] * 2
    divisors = [24024] * 2


class SavitzkyGolayDeriv1Cubic15(SavitzkyGolayNormalise):
    """Savitzky-Golay first cubic/quartic operator of size 15"""

    matrices = [[12922, -4121, -14150, -18334, -17842, -13843, -7506,
                 0,
                 7506, 13843, 17842, 18334, 14150, 4121, -12922]] * 2
    divisors = [334152] * 2


class SavitzkyGolayDeriv1Cubic17(SavitzkyGolay):
    """Savitzky-Golay first cubic/quartic operator of size 17"""

    matrices = [[748, -98, -643, -930, -1002, -902, -673, -358, 0, 358, 673, 902, 1002, 930, 643, 98, -748]] * 2
    divisors = [23256] * 2


class SavitzkyGolayDeriv1Cubic19(SavitzkyGolayNormalise):
    """Savitzky-Golay first cubic/quartic operator of size 19"""

    matrices = [[6936, 68, -4648, -7481, -8700, -8574, -7372, -5363, -2816,
                 0,
                 2816, 5363, 7372, 8574, 8700, 7481, 4648, -68, -6936]] * 2
    divisors = [255816] * 2


class SavitzkyGolayDeriv1Cubic21(SavitzkyGolayNormalise):
    """Savitzky-Golay first cubic/quartic operator of size 21"""

    matrices = [[84075, 10032, -43284, -78176, -96947, -101900, -95338, -79564, -56881, -29592,
                 0,
                 29592, 56881, 79564, 95338, 101900, 96947, 78176, 43284, -10032, -84075]] * 2
    divisors = [3634092] * 2


class SavitzkyGolayDeriv1Cubic23(SavitzkyGolayNormalise):
    """Savitzky-Golay first cubic/quartic operator of size 23"""

    matrices = [[3938, 815, -1518, -3140, -4130, -4567, -4530, -4098, -3350, -2365, -1222,
                 0,
                 1222, 2365, 3350, 4098, 4530, 4567, 4130, 3140, 1518, -815, -3938]] * 2
    divisors = [197340] * 2


class SavitzkyGolayDeriv1Cubic25(SavitzkyGolayNormalise):
    """Savitzky-Golay first cubic/quartic operator of size 25"""

    matrices = [[30866, 8602, -8525, -20982, -29236, -33754, -35003, -33450, -29562, -23806, -16649, -8558,
                 0,
                 8558, 16649, 23806, 29562, 33450, 35003, 33754, 29236, 20982, 8525, -8602, -30866]] * 2
    divisors = [1776060] * 2


class SavitzkyGolayDeriv1Quint7(SavitzkyGolay):
    """Savitzky-Golay first quintic/sextic derivative operator of size 7"""

    matrices = [[-1, 9, -45, 0, 45, -9, 1]] * 2
    divisors = [60] * 2


class SavitzkyGolayDeriv1Quint9(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 9"""

    matrices = [[-254, 1381, -2269, -2879, 0, 2879, 2269, -1381, 254]] * 2
    divisors = [8580] * 2


class SavitzkyGolayDeriv1Quint11(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 11"""

    matrices = [[-573, 2166, -1249, -3774, -3084, 0, 3084, 3774, 1249, -2166, 573]] * 2
    divisors = [17160] * 2


class SavitzkyGolayDeriv1Quint13(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 13"""

    matrices = [[-9647, 27093, -12, -33511, -45741, -31380, 0, 31380, 45741, 33511, 12, -27093, 9647]] * 2
    divisors = [291720] * 2


class SavitzkyGolayDeriv1Quint15(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 15"""

    matrices = [[-78351, 169819, 65229, -130506, -266401, -279975, -175125,
                 0,
                 175125, 279975, 266401, 130506, -65229, -169819, 78351]] * 2
    divisors = [2519400] * 2


class SavitzkyGolayDeriv1Quint17(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 17"""

    matrices = [[-14404, 24661, 16679, -8671, -32306, -43973, -40483, -23945,
                 0,
                 23945, 40483, 43973, 32306, 8671, -16679, -24661, 14404]] * 2
    divisors = [503880] * 2


class SavitzkyGolayDeriv1Quint19(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 19"""

    matrices = [[-255102, 349928, 322378, 9473, -348823, -604484, -686099, -583549, -332684,
                 0,
                 332684, 583549, 686099, 604484, 348823, -9473, -322378, -349928, 255102]] * 2
    divisors = [9806280] * 2


class SavitzkyGolayDeriv1Quint21(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 21"""

    matrices = [[
        -15033066, 16649358, 19052988, 6402438, -10949942, -26040033, -34807914, -35613829, -28754154, -15977364,
        0,
        15977364, 28754154, 35613829, 34807914, 26040033, 10949942, -6402438, -19052988, -16649358, 15033066
    ]] * 2
    divisors = [637408200] * 2


class SavitzkyGolayDeriv1Quint23(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 23"""

    matrices = [[
        -400653, 359157, 489687, 265164, -106911, -478349, -752859, -878634, -840937, -654687, -357045,
        0,
        357045, 654687, 840937, 878634, 752859, 478349, 106911, -265164, -489687, -359157, 400653
    ]] * 2
    divisors = [18747300] * 2


class SavitzkyGolayDeriv1Quint25(SavitzkyGolayNormalise):
    """Savitzky-Golay first quintic/sextic derivative operator of size 25"""

    matrices = [[
        -8322182, 6024183, 9604353, 6671883, 544668, -6301491, -12139321, -15896511, -17062146, -15593141, -11820675,
        -6356625, 0, 6356625, 11820675, 15593141, 17062146, 15896511, 12139321, 6301491, -544668, -6671883, -9604353,
        -6024183, 8322182
    ]] * 2
    divisors = [429214500] * 2


class SavitzkyGolayDeriv2Quad5(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 5"""

    matrices = [[2, -1, -2, -1, 2]] * 2
    divisors = [7] * 2


class SavitzkyGolayDeriv2Quad7(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 7"""

    matrices = [[5, 0, -3, -4, -3, 0, 5]] * 2
    divisors = [42] * 2


class SavitzkyGolayDeriv2Quad9(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 9"""

    matrices = [[28, 7, -8, -17, -20, -17, -8, 7, 28]] * 2
    divisors = [462] * 2


class SavitzkyGolayDeriv2Quad11(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 11"""

    matrices = [[15, 6, -1, -6, -9, -10, -9, -6, -1, 6, 15]] * 2
    divisors = [429] * 2


class SavitzkyGolayDeriv2Quad13(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 13"""

    matrices = [[22, 11, 2, -5, -10, -13, -14, -13, -10, -5, 2, 11, 22]] * 2
    divisors = [1001] * 2


class SavitzkyGolayDeriv2Quad15(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 15"""

    matrices = [[91, 52, 19, -8, -29, -44, -53, -56, -53, -44, -29, -8, 19, 52, 91]] * 2
    divisors = [6188] * 2


class SavitzkyGolayDeriv2Quad17(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 17"""

    matrices = [[40, 25, 12, 1, -8, -15, -20, -23, -24, -23, -20, -15, -8, 1, 12, 25, 40]] * 2
    divisors = [3876] * 2


class SavitzkyGolayDeriv2Quad19(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 19"""

    matrices = [[51, 34, 19, 6, -5, -14, -21, -26, -29, -30, -29, -26, -21, -14, -5, 6, 19, 34, 51]] * 2
    divisors = [6783] * 2


class SavitzkyGolayDeriv2Quad21(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 21"""

    matrices = [[
        190, 133, 82, 37, -2, -35, -62, -83, -98, -107, -110, -107, -98, -83, -62, -35, -2, 37, 82, 133, 190
    ]] * 2
    divisors = [33649] * 2


class SavitzkyGolayDeriv2Quad23(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 23"""

    matrices = [[
        77, 56, 37, 20, 5, -8, -19, -28, -35, -40, -43, -44, -43, -40, -35, -28, -19, -8, 5, 20, 37, 56, 77
    ]] * 2
    divisors = [17710] * 2


class SavitzkyGolayDeriv2Quad25(SavitzkyGolay):
    """Savitzky-Golay second quadratic/cubic derivative operator of size 25"""

    matrices = [[
        92, 69, 48, 29, 12, -3, -16, -27, -36, -43, -48, -51, -52, -51, -48, -43, -36, -27, -16, -3, 12, 29, 48, 69, 92
    ]] * 2
    divisors = [26910] * 2


class SavitzkyGolayDeriv2Quart7(SavitzkyGolay):
    """Savitzky-Golay second quartic/quintic derivative operator of size 7"""

    matrices = [[-13, 67, -19, -70, -19, 67, -13]] * 2
    divisors = [132] * 2


class SavitzkyGolayDeriv2Quart9(SavitzkyGolay):
    """Savitzky-Golay second quartic/quintic derivative operator of size 9"""

    matrices = [[-126, 371, 151, -211, -370, -211, 151, 371, -126]] * 2
    divisors = [1716] * 2


class SavitzkyGolayDeriv2Quart11(SavitzkyGolay):
    """Savitzky-Golay second quartic/quintic derivative operator of size 11"""

    matrices = [[-90, 174, 146, 1, -136, -190, -136, 1, 146, 174, -90]] * 2
    divisors = [1716] * 2


class SavitzkyGolayDeriv2Quart13(SavitzkyGolayNormalise):
    """Savitzky-Golay second quartic/quintic derivative operator of size 13"""

    matrices = [[-2211, 2970, 3504, 1614, -971, -3016, -3780, -3016, -971, 1614, 3504, 2970, -2211]] * 2
    divisors = [58344] * 2


class SavitzkyGolayDeriv2Quart15(SavitzkyGolayNormalise):
    """Savitzky-Golay second quartic/quintic derivative operator of size 15"""

    matrices = [[-31031, 29601, 44495, 31856, 6579, -19751, -38859, -45780, -38859, -19751,
                 6579, 31856, 44495, 29601, -31031]] * 2
    divisors = [1108536] * 2


class SavitzkyGolayDeriv2Quart17(SavitzkyGolayNormalise):
    """Savitzky-Golay second quartic/quintic derivative operator of size 17"""

    matrices = [[
        -2132, 1443, 2691, 2405, 1256, -207, -1557, -2489, -2820, -2489, -1557, -207, 1256, 2405, 2691, 1443, -2132
    ]] * 2
    divisors = [100776] * 2


class SavitzkyGolayDeriv2Quart19(SavitzkyGolayNormalise):
    """Savitzky-Golay second quartic/quintic derivative operator of size 19"""

    matrices = [[
        -32028, 15028, 35148, 36357, 25610, 8792, -9282, -24867, -35288,
        -38940,
        -35288, -24867, -9282, 8792, 25610, 36357, 35148, 15028, -32028
    ]] * 2
    divisors = [1961256] * 2


class SavitzkyGolayDeriv2Quart21(SavitzkyGolayNormalise):
    """Savitzky-Golay second quartic/quintic derivative operator of size 21"""

    matrices = [[
        -12597, 3876, 11934, 13804, 11451, 6578, 626, -5226, -10061, -13224,
        -14322,
        -13224, -10061, -5226, 626, 6578, 11451, 13804, 11934, 3876, -12597
    ]] * 2
    divisors = [980628] * 2


class SavitzkyGolayDeriv2Quart23(SavitzkyGolayNormalise):
    """Savitzky-Golay second quartic/quintic derivative operator of size 23"""

    matrices = [[
        -115577, 20615, 93993, 119510, 110545, 78903, 34815, -13062, -57645, -93425, -116467,
        -124410,
        -116467, -93425, -57645, -13062, 34815, 78903, 110545, 119510, 93993, 20615, -115577
    ]] * 2
    divisors = [11248380] * 2


class SavitzkyGolayDeriv2Quart25(SavitzkyGolayNormalise):
    """Savitzky-Golay second quartic/quintic derivativeoperator of size 25"""

    matrices = [[
        -143198, 10373, 99385, 137803, 138262, 112067, 69193, 18285, -33342, -79703, -116143, -139337,
        -147290,
        -139337, -116143, -79703, -33342, 18285, 69193, 112067, 138262, 137803, 99385, 10373, -143198
    ]] * 2
    divisors = [17168580] * 2


class SavitzkyGolayDeriv3Cub5(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 5"""

    matrices = [[-1, 2, 0, -2, 1]] * 2
    divisors = [2] * 2


class SavitzkyGolayDeriv3Cub7(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 7"""

    matrices = [[-1, 1, 1, 0, -1, -1, 1]] * 2
    divisors = [6] * 2


class SavitzkyGolayDeriv3Cub9(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 9"""

    matrices = [[-14, 7, 13, 9, 0, -9, -13, -7, 14]] * 2
    divisors = [198] * 2


class SavitzkyGolayDeriv3Cub11(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 11"""

    matrices = [[-30, 6, 22, 23, 14, 0, -14, -23, -22, -6, 30]] * 2
    divisors = [858] * 2


class SavitzkyGolayDeriv3Cub13(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 13"""

    matrices = [[-11, 0, 6, 8, 7, 4, 0, -4, -7, -8, -6, 0, 11]] * 2
    divisors = [572] * 2


class SavitzkyGolayDeriv3Cub15(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 15"""

    matrices = [[-91, -13, 35, 58, 61, 49, 27, 0, -27, -49, -61, -58, -35, 13, 91]] * 2
    divisors = [7956] * 2


class SavitzkyGolayDeriv3Cub17(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 17"""

    matrices = [[-28, -7, 7, 15, 18, 17, 13, 7, 0, -7, -13, -17, -18, -15, -7, 7, 28]] * 2
    divisors = [3876] * 2


class SavitzkyGolayDeriv3Cub19(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 19"""

    matrices = [[-204, -68, 28, 89, 120, 126, 112, 83, 44, 0, -44, -83, -112, -126, -120, -89, -28, 68, 204]] * 2
    divisors = [42636] * 2


class SavitzkyGolayDeriv3Cub21(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 21"""

    matrices = [[-285, -114, 12, 98, 149, 170, 166, 142, 103, 54,
                 0, -54, -103, -142, -166, -170, -149, -98, -12, 114, 285]] * 2
    divisors = [86526] * 2


class SavitzkyGolayDeriv3Cub23(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 23"""

    matrices = [[-77, -35, -3, 20, 35, 43, 45, 42, 35, 25, 13,
                 0, -13, -25, -35, -42, -45, -43, -35, -20, 3, 35, 77]] * 2
    divisors = [32890] * 2


class SavitzkyGolayDeriv3Cub25(SavitzkyGolay):
    """Savitzky-Golay third cubic/quartic derivative operator of size 25"""

    matrices = [[
        -506, -253, -55, 93, 196, 259, 287, 285, 258, 211, 149, 77,
        0,
        -77, -149, -211, -258, -285, -287, -259, -196, -93, 55, 253, 506
    ]] * 2
    divisors = [296010] * 2


class SavitzkyGolayDeriv3Quint7(SavitzkyGolay):
    """Savitzky-Golay third quintic/sexic derivative operator of size 7"""

    matrices = [[1, -8, 13, 0, -13, 8, -1]] * 2
    divisors = [8] * 2


class SavitzkyGolayDeriv3Quint9(SavitzkyGolay):
    """Savitzky-Golay third quintic/sexic derivative operator of size 9"""

    matrices = [[100, -457, 256, 459, 0, -459, -256, 457, -100]] * 2
    divisors = [1144] * 2


class SavitzkyGolayDeriv3Quint11(SavitzkyGolay):
    """Savitzky-Golay third quintic/sexic derivative operator of size 11"""

    matrices = [[129, -402, -11, 340, 316, 0, -316, -340, 11, 402, -129]] * 2
    divisors = [2288] * 2


class SavitzkyGolayDeriv3Quint13(SavitzkyGolayNormalise):
    """Savitzky-Golay third quintic/sexic derivative operator of size 13"""

    matrices = [[1430, -3267, -1374, 1633, 3050, 2252, 0, -2252, -3050, -1633, 1374, 3267, -1430]] * 2
    divisors = [38896] * 2


class SavitzkyGolayDeriv3Quint15(SavitzkyGolayNormalise):
    """Savitzky-Golay third quintic/sexic derivative operator of size 15"""

    matrices = [[8281, -14404, -10379, 1916, 11671, 14180, 9315,
                 0, -9315, -14180, -11671, -1916, 10379, 14404, -8281]] * 2
    divisors = [335920] * 2


class SavitzkyGolayDeriv3Quint17(SavitzkyGolayNormalise):
    """Savitzky-Golay third quintic/sexic derivative operator of size 17"""

    matrices = [[1144, -1547, -1508, -351, 876, 1595, 1604, 983,
                 0, -983, -1604, -1595, -876, 351, 1508, 1547, -1144]] * 2
    divisors = [67184] * 2


class SavitzkyGolayDeriv3Quint19(SavitzkyGolayNormalise):
    """Savitzky-Golay third quintic/sexic derivative operator of size 19"""

    matrices = [[
        15810, -16796, -20342, -9818, 4329, 15546, 20525, 18554, 10868,
        0,
        -10868, -18554, -20525, -15546, -4329, 9818, 20342, 16796, -15810
    ]] * 2
    divisors = [1307504] * 2


class SavitzkyGolayDeriv3Quint21(SavitzkyGolayNormalise):
    """Savitzky-Golay third quintic/sexic derivative operator of size 21"""

    matrices = [[
        748068, -625974, -908004, -598094, -62644, 448909, 787382, 887137, 749372, 425412,
        0,
        -425412, -749372, -887137, -787382, -448909, 62644, 598094, 908004, 625974, -748068
    ]] * 2
    divisors = [84987760] * 2


class SavitzkyGolayDeriv3Quint23(SavitzkyGolayNormalise):
    """Savitzky-Golay third quintic/sexic derivative operator of size 23"""

    matrices = [[
        49115, -32224, -55233, -43928, -16583, 13632, 38013, 51684, 52959, 42704, 23699,
        0,
        -23699, -42704, -52959, -51684, -38013, -13632, 16583, 43928, 55233, 32224, -49115
    ]] * 2
    divisors = [7498920] * 2


class SavitzkyGolayDeriv3Quint25(SavitzkyGolayNormalise):
    """Savitzky-Golay third quintic/sexic derivative operator of size 25"""

    matrices = [[
        284372, -144463, -293128, -266403, -146408, 5131, 144616, 244311, 290076, 279101, 217640, 118745,
        0,
        -118745, -217640, -279101, -290076, -244311, -144616, -5131, 146408, 266403, 293128, 144463, -284372
    ]] * 2
    divisors = [57228600] * 2


class SavitzkyGolayDeriv4Quart7(SavitzkyGolay):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 7"""

    matrices = [[3, -7, 1, 6, 1, -7, 3]] * 2
    divisors = [11] * 2


class SavitzkyGolayDeriv4Quart9(SavitzkyGolay):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 9"""

    matrices = [[14, -21, -11, 9, 18, 9, -11, -21, 14]] * 2
    divisors = [143] * 2


class SavitzkyGolayDeriv4Quart11(SavitzkyGolay):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 11"""

    matrices = [[6, -6, -6, -1, 4, 6, 4, -1, -6, -6, 6]] * 2
    divisors = [143] * 2


class SavitzkyGolayDeriv4Quart13(SavitzkyGolay):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 13"""

    matrices = [[99, -66, -96, -54, 11, 64, 84, 64, 11, -54, -96, -66, 99]] * 2
    divisors = [4862] * 2


class SavitzkyGolayDeriv4Quart15(SavitzkyGolay):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 15"""

    matrices = [[1001, -429, -869, -704, -249, 251, 621, 756, 621, 251, -249, -704, -869, -429, 1001]] * 2
    divisors = [92378] * 2


class SavitzkyGolayDeriv4Quart17(SavitzkyGolay):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 17"""

    matrices = [[52, -13, -39, -39, -24, -3, 17, 31, 36, 31, 17, -3, -24, -39, -39, -13, 52]] * 2
    divisors = [8398] * 2


class SavitzkyGolayDeriv4Quart19(SavitzkyGolay):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 19"""

    matrices = [[612, -68, -388, -453, -354, -168, 42, 227, 352,
                 396, 352, 227, 42, -168, -354, -453, -388, -68, 612]] * 2
    divisors = [163438] * 2


class SavitzkyGolayDeriv4Quart21(SavitzkyGolay):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 21"""

    matrices = [[969, 0, -510, -680, -615, -406, -130, 150, 385, 540,
                 594, 540, 385, 150, -130, -406, -615, -680, -510, 0, 969]] * 2
    divisors = [408595] * 2


class SavitzkyGolayDeriv4Quart23(SavitzkyGolayNormalise):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 23"""

    matrices = [[
        1463, 133, -627, -950, -955, -747, -417, -42, 315, 605, 793,
        858,
        793, 605, 315, -42, -417, -747, -955, -950, -627, 133, 1463
    ]] * 2
    divisors = [937365] * 2


class SavitzkyGolayDeriv4Quart25(SavitzkyGolayNormalise):
    """Savitzky-Golay fourth quartic/quintic derivative operator of size 25"""

    matrices = [[
        1518, 253, -517, -897, -982, -857, -597, -267, 78, 393, 643, 803,
        858,
        803, 643, 393, 78, -267, -597, -857, -982, -897, -517, 253, 1518
    ]] * 2
    divisors = [1430715] * 2


class SavitzkyGolayDeriv5Quint7(SavitzkyGolay):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 7"""

    matrices = [[-1, 4, -5, 0, 5, -4, 1]] * 2
    divisors = [2] * 2


class SavitzkyGolayDeriv5Quint9(SavitzkyGolay):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 9"""

    matrices = [[-4, 11, -4, -9, 0, 9, 4, -11, 4]] * 2
    divisors = [26] * 2


class SavitzkyGolayDeriv5Quint11(SavitzkyGolay):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 11"""

    matrices = [[-3, 6, 1, -4, -4, 0, 4, 4, -1, -6, 3]] * 2
    divisors = [52] * 2


class SavitzkyGolayDeriv5Quint13(SavitzkyGolay):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 13"""

    matrices = [[-22, 33, 18, -11, -26, -20, 0, 20, 26, 11, -18, -33, 22]] * 2
    divisors = [884] * 2


class SavitzkyGolayDeriv5Quint15(SavitzkyGolayNormalise):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 15"""

    matrices = [[-1001, 1144, 979, 44, -751, -1000, -675, 0, 675, 1000, 751, -44, -979, -1144, 1001]] * 2
    divisors = [83980] * 2


class SavitzkyGolayDeriv5Quint17(SavitzkyGolay):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 17"""

    matrices = [[-104, 91, 104, 39, -36, -83, -88, -55, 0, 55, 88, 83, 36, -39, -104, -91, 104]] * 2
    divisors = [16796] * 2


class SavitzkyGolayDeriv5Quint19(SavitzkyGolay):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 19"""

    matrices = [[-102, 68, 98, 58, -3, -54, -79, -74, -44, 0, 44, 74, 79, 54, 3, -58, -98, -68, 102]] * 2
    divisors = [29716] * 2


class SavitzkyGolayDeriv5Quint21(SavitzkyGolayNormalise):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 21"""

    matrices = [[
        -3876, 1938, 3468, 2618, 788, -1063, -2354, -2819, -2444, -1404,
        0,
        1404, 2444, 2819, 2354, 1063, -788, -2618, -3468, -1938, 3876
    ]] * 2
    divisors = [1931540] * 2


class SavitzkyGolayDeriv5Quint23(SavitzkyGolay):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 23"""

    matrices = [[
        -209, 76, 171, 152, 77, -12, -87, -132, -141, -116, -65,
        0,
        65, 116, 141, 132, 87, 12, -77, -152, -171, -76, 209
    ]] * 2
    divisors = [170430] * 2


class SavitzkyGolayDeriv5Quint25(SavitzkyGolay):
    """Savitzky-Golay fifth quintic/sexic derivative operator of size 25"""

    matrices = [[
        -1012, 253, 748, 753, 488, 119, -236, -501, -636, -631, -500, -275,
        0,
        275, 500, 631, 636, 501, 236, -119, -488, -753, -748, -253, 1012
    ]] * 2
    divisors = [1300650] * 2



from __future__ import annotations

from vsexprtools import ExprOp
from vsrgtools import RemoveGrainMode, RemoveGrainModeT, bilateral, gauss_blur, removegrain
from vstools import check_variable, get_y, plane, vs

from .edge import Kirsch, MinMax, Prewitt, PrewittTCanny
from .masks import range_mask
from .morpho import Morpho
from .types import GenericMaskT
from .utils import normalize_mask

__all__ = [
    'detail_mask',
    'detail_mask_neo',
    'simple_detail_mask',
    'multi_detail_mask',
]


def detail_mask(
    clip: vs.VideoNode, brz_mm: float, brz_ed: float,
    minmax: MinMax = MinMax(rady=3, radc=2),
    edge: GenericMaskT = Kirsch
) -> vs.VideoNode:
    assert check_variable(clip, detail_mask)

    range_mask = Morpho.binarize(minmax.edgemask(clip), brz_mm)

    edges = Morpho.binarize(normalize_mask(edge, clip), brz_ed)

    mask = ExprOp.MAX.combine(range_mask, edges)

    mask = removegrain(mask, 22)
    mask = removegrain(mask, 11)

    return mask.std.Limiter()


def detail_mask_neo(
    clip: vs.VideoNode, sigma: float = 1.0, detail_brz: float = 0.05, lines_brz: float = 0.08,
    edgemask: GenericMaskT = Prewitt, rg_mode: RemoveGrainModeT = RemoveGrainMode.MINMAX_MEDIAN_OPP
) -> vs.VideoNode:
    assert check_variable(clip, "detail_mask_neo")

    clip_y = get_y(clip)
    blur_pf = gauss_blur(clip_y, sigma * 0.75)

    blur_pref = bilateral(clip_y, sigma, ref=blur_pf)
    blur_pref_diff = ExprOp.SUB.combine(blur_pref, clip_y).std.Deflate()
    blur_pref = Morpho.inflate(blur_pref_diff, iterations=4)

    prew_mask = normalize_mask(edgemask, clip_y).std.Deflate().std.Inflate()

    if detail_brz > 0:
        blur_pref = Morpho.binarize(blur_pref, detail_brz)

    if lines_brz > 0:
        prew_mask = Morpho.binarize(prew_mask, lines_brz)

    merged = ExprOp.ADD.combine(blur_pref, prew_mask)

    return removegrain(merged, rg_mode).std.Limiter()


def simple_detail_mask(
    clip: vs.VideoNode, sigma: float | None = None, rad: int = 3, brz_a: float = 0.025, brz_b: float = 0.045
) -> vs.VideoNode:
    y = plane(clip, 0)

    blur = gauss_blur(y, sigma) if sigma else y

    mask_a = Morpho.binarize(range_mask(blur, rad=rad), brz_a)

    mask_b = Morpho.binarize(PrewittTCanny.edgemask(blur), brz_b)

    mask = ExprOp.MAX.combine(mask_a, mask_b)

    return removegrain(removegrain(mask, 22), 11).std.Limiter()


def multi_detail_mask(clip: vs.VideoNode, thr: float = 0.015) -> vs.VideoNode:
    general_mask = simple_detail_mask(clip, rad=1, brz_a=1, brz_b=24.3 * thr)

    return ExprOp.MIN.combine(
        ExprOp.MIN.combine(
            simple_detail_mask(clip, brz_a=1, brz_b=2 * thr),
            Morpho.maximum(general_mask, iterations=4) .std.Inflate()
        ), general_mask.std.Maximum()
    )



from __future__ import annotations

from typing import Sequence

from vsexprtools import ExprOp, ExprVars, aka_expr_available, norm_expr
from vsrgtools import gauss_blur
from vstools import (
    ColorRange, CustomRuntimeError, FuncExceptT, StrList, check_variable, get_lowest_value, get_peak_value,
    get_sample_type, get_y, plane, vs, core
)

__all__ = [
    'adg_mask',
    'retinex',
]


def adg_mask(
    clip: vs.VideoNode, luma_scaling: float = 8.0, relative: bool = False, func: FuncExceptT | None = None
) -> vs.VideoNode:
    func = func or adg_mask

    assert check_variable(clip, func)

    y = plane(clip, 0).std.PlaneStats(prop='P')

    if not aka_expr_available:
        if relative:
            raise CustomRuntimeError(
                "You don't have akarin plugin, you can't use this function!", func, 'relative=True'
            )

        return y.adg.Mask(luma_scaling)  # type: ignore

    assert y.format

    peak = get_peak_value(y)

    is_integer = y.format.sample_type == vs.INTEGER

    x_string, aft_int = (f'x {peak} / ', f' {peak} * 0.5 +') if is_integer else ('x ', '')

    if relative:
        x_string += 'Y! Y@ 0.5 < x.PMin 0 max 0.5 / log Y@ * x.PMax 1.0 min 0.5 / log Y@ * ? '

    x_string += '0 0.999 clamp X!'

    return norm_expr(
        y, f'{x_string} 1 X@ X@ X@ X@ X@ '
        '18.188 * 45.47 - * 36.624 + * 9.466 - * 1.124 + * - '
        f'x.PAverage 2 pow {luma_scaling} * pow {aft_int}'
    )


def retinex(
    clip: vs.VideoNode, sigma: Sequence[float] = [25, 80, 250],
    lower_thr: float = 0.001, upper_thr: float = 0.001,
    fast: bool = False, func: FuncExceptT | None = None
) -> vs.VideoNode:
    func = func or retinex

    assert check_variable(clip, func)

    sigma = sorted(sigma)

    y = get_y(clip)

    if not aka_expr_available or not hasattr(core, 'psm'):
        if fast:
            raise CustomRuntimeError(
                "You don't have {missing} plugin, you can't use this function!", func, 'fast=True',
                missing=iter(x for x in ('akarin', 'psm') if not hasattr(core, x))
            )

        return y.retinex.MSRCP(sigma, lower_thr, upper_thr)  # type: ignore

    y = y.std.PlaneStats()
    is_float = get_sample_type(y) is vs.FLOAT

    if is_float:
        luma_float = norm_expr(y, "x x.PlaneStatsMin - x.PlaneStatsMax x.PlaneStatsMin - /")
    else:
        luma_float = norm_expr(y, "1 x.PlaneStatsMax x.PlaneStatsMin - / x x.PlaneStatsMin - *", None, vs.GRAYS)

    slen, slenm = len(sigma), len(sigma) - 1

    expr_msr = StrList([
        f"{x} 0 <= 1 x {x} / 1 + ? "
        for x in ExprVars(1, slen + (not fast))
    ])

    if fast:
        expr_msr.append("x.PlaneStatsMax 0 <= 1 x x.PlaneStatsMax / 1 + ? ")
        sigma = sigma[:-1]

    expr_msr.extend(ExprOp.ADD * slenm)
    expr_msr.append(f"log {slen} /")

    msr = norm_expr([luma_float, (gauss_blur(luma_float, i) for i in sigma)], expr_msr)

    msr_stats = msr.psm.PlaneMinMax(lower_thr, upper_thr)

    expr_balance = "x x.psmMin - x.psmMax x.psmMin - /"

    if not is_float:
        expr_balance = f"{expr_balance} {{ymax}} {{ymin}} - * {{ymin}} + round {{ymin}} {{ymax}} clamp"

    return norm_expr(
        msr_stats, expr_balance, None, y,
        ymin=get_lowest_value(y, False, ColorRange.LIMITED),
        ymax=get_peak_value(y, False, ColorRange.LIMITED)
    )



from __future__ import annotations

from functools import partial
from typing import Any, Sequence

from vsexprtools import ExprOp, ExprToken, norm_expr
from vsrgtools import gauss_blur
from vsrgtools.util import wmean_matrix
from vstools import check_variable, core, get_peak_value, get_y, iterate, plane, scale_value, vs

from .details import multi_detail_mask
from .edge import FDoGTCanny, Kirsch, Prewitt
from .funcs import retinex
from .morpho import Morpho
from .types import Coordinates, GenericMaskT
from .utils import normalize_mask

__all__ = [
    'ringing_mask',

    'luma_mask', 'luma_credit_mask',

    'tcanny_retinex',

    'limited_linemask'
]


def ringing_mask(
    clip: vs.VideoNode,
    rad: int = 2, brz: float = 0.35,
    thmi: float = 0.315, thma: float = 0.5,
    thlimi: float = 0.195, thlima: float = 0.392,
    credit_mask: GenericMaskT = Prewitt, **kwargs: Any
) -> vs.VideoNode:
    assert check_variable(clip, ringing_mask)

    thmi, thma, thlimi, thlima = (
        scale_value(t, 32, clip) for t in [thmi, thma, thlimi, thlima]
    )

    edgemask = normalize_mask(credit_mask, plane(clip, 0), **kwargs).std.Limiter()

    light = norm_expr(edgemask, f'x {thlimi} - {thma - thmi} / {ExprToken.RangeMax} *')

    shrink = Morpho.dilation(light, rad)
    shrink = Morpho.binarize(shrink, brz)
    shrink = Morpho.erosion(shrink, 2)
    shrink = iterate(shrink, partial(core.std.Convolution, matrix=wmean_matrix), 2)

    strong = norm_expr(edgemask, f'x {thmi} - {thlima - thlimi} / {ExprToken.RangeMax} *')
    expand = Morpho.dilation(strong, rad)

    mask = norm_expr([expand, strong, shrink], 'x y z max -')

    return ExprOp.convolution('x', wmean_matrix, premultiply=2, multiply=2, clamp=True)(mask)


def luma_mask(clip: vs.VideoNode, thr_lo: float, thr_hi: float, invert: bool = True) -> vs.VideoNode:
    peak = get_peak_value(clip)

    lo, hi = (peak, 0) if invert else (0, peak)
    inv_pre, inv_post = (peak, '-') if invert else ('', '')

    thr_lo, thr_hi = scale_value(thr_lo, 32, clip), scale_value(thr_hi, 32, clip)

    return norm_expr(
        get_y(clip),
        f'x {thr_lo} < {lo} x {thr_hi} > {hi} {inv_pre} x {thr_lo} - {thr_lo} {thr_hi} - / {peak} * {inv_post} ? ?'
    )


def luma_credit_mask(
    clip: vs.VideoNode, thr: float = 0.9, edgemask: GenericMaskT = FDoGTCanny, draft: bool = False, **kwargs: Any
) -> vs.VideoNode:
    y = plane(clip, 0)

    edge_mask = normalize_mask(edgemask, y, **kwargs)

    credit_mask = norm_expr([edge_mask, y], f'y {scale_value(thr, 32, y)} > y 0 ? x min')

    if not draft:
        credit_mask = Morpho.maximum(credit_mask, iterations=4)
        credit_mask = Morpho.inflate(credit_mask, iterations=2)

    return credit_mask


def tcanny_retinex(
    clip: vs.VideoNode, thr: float, sigma: Sequence[float] = [50, 200, 350], blur_sigma: float = 1.0
) -> vs.VideoNode:
    blur = gauss_blur(clip, blur_sigma)

    msrcp = retinex(blur, sigma, upper_thr=thr, fast=True, func=tcanny_retinex)

    tcunnied = msrcp.tcanny.TCanny(mode=1, sigma=1)

    return Morpho.minimum(tcunnied, None, Coordinates.CORNERS)


def limited_linemask(
    clip: vs.VideoNode,
    sigmas: list[float] = [0.000125, 0.0025, 0.0055],
    detail_sigmas: list[float] = [0.011, 0.013],
    edgemasks: Sequence[GenericMaskT] = [Kirsch],
    **kwargs: Any
) -> vs.VideoNode:
    clip_y = plane(clip, 0)

    return ExprOp.ADD(
        (normalize_mask(edge, clip_y, **kwargs) for edge in edgemasks),
        (tcanny_retinex(clip_y, s) for s in sigmas),
        (multi_detail_mask(clip_y, s) for s in detail_sigmas)
    )



from __future__ import annotations

from dataclasses import dataclass
from itertools import zip_longest
from math import floor, sqrt
from typing import Any, Literal, Sequence, Tuple

from vsexprtools import ExprList, ExprOp, ExprToken, aka_expr_available, norm_expr
from vsrgtools.util import wmean_matrix
from vstools import (
    ConvMode, CustomIndexError, FuncExceptT, PlanesT, StrList, check_variable, copy_signature, core, fallback,
    inject_self, interleave_arr, iterate, scale_value, to_arr, vs
)

from .types import Coordinates, MorphoFunc, XxpandMode

__all__ = [
    'Morpho', 'CoordsT',
    'grow_mask'
]

CoordsT = int | Tuple[int, ConvMode] | Sequence[int]


def _minmax_method(  # type: ignore
    self: Morpho, src: vs.VideoNode, thr: float | None = None,
    coords: CoordsT | None = [1] * 8,
    iterations: int = 1, multiply: float | None = None, planes: PlanesT = None,
    *, func: FuncExceptT | None = None, **kwargs: Any
) -> vs.VideoNode:
    ...


def _morpho_method(  # type: ignore
    self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
    coords: CoordsT = 5, multiply: float | None = None,
    *, func: FuncExceptT | None = None, **kwargs: Any
) -> vs.VideoNode:
    ...


def _morpho_method2(  # type: ignore
    self: Morpho, clip: vs.VideoNode, sw: int, sh: int | None = None, mode: XxpandMode = XxpandMode.RECTANGLE,
    thr: float | None = None, planes: PlanesT = None, *, func: FuncExceptT | None = None, **kwargs: Any
) -> vs.VideoNode:
    ...


@dataclass
class Morpho:
    planes: PlanesT = None
    func: FuncExceptT | None = None
    fast: bool | None = None

    def __post_init__(self) -> None:
        self._fast = fallback(self.fast, aka_expr_available) and aka_expr_available

    def _check_params(
        self, radius: int, thr: float | None, coords: CoordsT, planes: PlanesT, func: FuncExceptT
    ) -> tuple[FuncExceptT, PlanesT]:
        if radius < 1:
            raise CustomIndexError('radius has to be greater than 0!', func, radius)

        if isinstance(coords, (int, tuple)):
            size = coords if isinstance(coords, int) else coords[0]

            if size < 2:
                raise CustomIndexError('when int or tuple, coords has to be greater than 1!', func, coords)

            if not self._fast and size != 3:
                raise CustomIndexError(
                    'with fast=False or no akarin plugin, you must have coords=3!', func, coords
                )
        elif len(coords) != 8:
            raise CustomIndexError('when a list, coords must contain exactly 8 numbers!', func, coords)

        if thr is not None and thr < 0.0:
            raise CustomIndexError('thr must be a positive number!', func, coords)

        return self.func or func, self.planes if planes is None else planes

    @classmethod
    def _morpho_xx_imum(
        cls, src: vs.VideoNode, thr: float | None, op: Literal[ExprOp.MIN, ExprOp.MAX],
        coords: CoordsT, multiply: float | None = None, clamp: bool = False
    ) -> StrList:
        exclude = list[tuple[int, int]]()

        if isinstance(coords, (int, tuple)):
            if isinstance(coords, tuple):
                size, mode = coords
            else:
                size, mode = coords, ConvMode.SQUARE

            assert size > 1

            radius = size // 2

            if size % 2 == 0:
                exclude.extend((x, radius) for x in range(-radius, radius + 1))
                exclude.append((radius, radius - 1))
        else:
            coords = list(coords)
            coords.insert(len(coords) // 2, 1)
            radius, mode = floor(sqrt(len(coords)) / 2), ConvMode.SQUARE

        matrix = ExprOp.matrix('x', radius, mode, exclude)

        if not isinstance(coords, (int, tuple)):
            matrix = ExprList([x for x, coord in zip(matrix, coords) if coord])

        matrix = ExprList(interleave_arr(matrix, op * matrix.mlength, 2))

        if thr is not None:
            matrix.append('x', scale_value(thr, 32, src), ExprOp.SUB, ExprOp.MAX)

        if multiply is not None:
            matrix.append(multiply, ExprOp.MUL)

        if clamp:
            matrix.append(ExprOp.clamp())

        return matrix

    def _mm_func(
        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
        coords: CoordsT = 5, multiply: float | None = None, *, func: FuncExceptT,
        mm_func: MorphoFunc, op: Literal[ExprOp.MIN, ExprOp.MAX], **kwargs: Any
    ) -> vs.VideoNode:
        func, planes = self._check_params(radius, thr, coords, planes, func)

        if self._fast:
            mm_func = norm_expr  # type: ignore[assignment]
            kwargs.update(expr=self._morpho_xx_imum(src, thr, op, coords, multiply))
        elif isinstance(coords, (int, tuple)):
            if isinstance(coords, tuple):
                if coords[1] is not ConvMode.SQUARE:
                    raise CustomIndexError(
                        'with fast=False or no akarin plugin, you must have ConvMode.SQUARE!', func, coords
                    )

                coords = coords[0]

            if coords != 3:
                raise CustomIndexError(
                    'with fast=False or no akarin plugin, you must have coords=3!', func, coords
                )

            kwargs.update(coordinates=[1] * 8)

        if not self._fast:
            if thr is not None:
                kwargs.update(threshold=scale_value(thr, 32, src))

            if multiply is not None:
                orig_mm_func = mm_func

                @copy_signature(mm_func)
                def _mm_func(*args: Any, **kwargs: Any) -> Any:
                    return orig_mm_func(*args, **kwargs).std.Expr(f'x {multiply} *')

                mm_func = _mm_func

        return iterate(src, mm_func, radius, planes=planes, **kwargs)

    @inject_self
    def xxpand_transform(
        self, clip: vs.VideoNode, op: Literal[ExprOp.MIN, ExprOp.MAX], sw: int, sh: int | None = None,
        mode: XxpandMode = XxpandMode.RECTANGLE, thr: float | None = None,
        planes: PlanesT = None, *, func: FuncExceptT | None = None
    ) -> vs.VideoNode:
        func, planes = self._check_params(1, thr, 3, planes, func or self.xxpand_transform)

        sh = fallback(sh, sw)

        if op not in {ExprOp.MIN, ExprOp.MAX}:
            raise NotImplementedError

        function = self.maximum if op is ExprOp.MAX else self.minimum

        for wi, hi in zip_longest(range(sw, -1, -1), range(sh, -1, -1), fillvalue=0):
            if wi > 0 and hi > 0:
                coords = Coordinates.from_xxpand_mode(mode, wi)
            elif wi > 0:
                coords = Coordinates.HORIZONTAL
            elif hi > 0:
                coords = Coordinates.VERTICAL
            else:
                break

            clip = function(clip, thr, coords, planes=planes, func=func)

        return clip

    def _xxflate(
        self: Morpho, inflate: bool, src: vs.VideoNode, radius: int, planes: PlanesT, thr: float | None,
        multiply: float | None, *, func: FuncExceptT
    ) -> vs.VideoNode:
        assert check_variable(src, func)

        expr = ExprOp.matrix('x', radius, exclude=[(0, 0)])

        conv_len = len(expr)

        expr.append(ExprOp.ADD * expr.mlength)

        if src.format.sample_type is vs.INTEGER:
            expr.append(radius * 4, ExprOp.ADD)

        expr.append(conv_len, ExprOp.DIV)
        expr.append('x', ExprOp.MAX if inflate else ExprOp.MIN)

        if thr is not None:
            thr = scale_value(thr, 32, src)
            limit = ['x', thr, ExprOp.ADD] if inflate else ['x', thr, ExprOp.SUB, ExprToken.RangeMin, ExprOp.MAX]

            expr.append(limit, ExprOp.MIN if inflate else ExprOp.MAX)

        if multiply is not None:
            expr.append(multiply, ExprOp.MUL)

        return norm_expr(src, expr, planes)

    @inject_self
    @copy_signature(_minmax_method)
    def maximum(
        self, src: vs.VideoNode, thr: float | None = None, coords: CoordsT | None = None,
        iterations: int = 1, multiply: float | None = None, planes: PlanesT = None,
        *, func: FuncExceptT | None = None, **kwargs: Any
    ) -> vs.VideoNode:
        return self.dilation(src, iterations, planes, thr, coords or ([1] * 8), multiply, func=func, **kwargs)

    @inject_self
    @copy_signature(_minmax_method)
    def minimum(
        self, src: vs.VideoNode, thr: float | None = None, coords: CoordsT | None = None,
        iterations: int = 1, multiply: float | None = None, planes: PlanesT = None,
        *, func: FuncExceptT | None = None, **kwargs: Any
    ) -> vs.VideoNode:
        return self.erosion(src, iterations, planes, thr, coords or ([1] * 8), multiply, func=func, **kwargs)

    @inject_self
    def inflate(
        self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
        iterations: int = 1, multiply: float | None = None, *, func: FuncExceptT | None = None
    ) -> vs.VideoNode:
        for _ in range(iterations):
            src = self._xxflate(True, src, radius, planes, thr, multiply, func=func or self.inflate)
        return src

    @inject_self
    def deflate(
        self: Morpho, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
        iterations: int = 1, multiply: float | None = None, *, func: FuncExceptT | None = None
    ) -> vs.VideoNode:
        for _ in range(iterations):
            src = self._xxflate(False, src, radius, planes, thr, multiply, func=func or self.deflate)
        return src

    @inject_self
    @copy_signature(_morpho_method)
    def dilation(self, *args: Any, func: FuncExceptT | None = None, **kwargs: Any) -> vs.VideoNode:
        return self._mm_func(*args, func=func or self.dilation, mm_func=core.std.Maximum, op=ExprOp.MAX, **kwargs)

    @inject_self
    @copy_signature(_morpho_method)
    def erosion(self, *args: Any, func: FuncExceptT | None = None, **kwargs: Any) -> vs.VideoNode:
        return self._mm_func(*args, func=func or self.erosion, mm_func=core.std.Minimum, op=ExprOp.MIN, **kwargs)

    @inject_self
    @copy_signature(_morpho_method2)
    def expand(self, clip: vs.VideoNode, *args: Any, func: FuncExceptT | None = None, **kwargs: Any) -> vs.VideoNode:
        return self.xxpand_transform(clip, ExprOp.MAX, *args, func=func, **kwargs)

    @inject_self
    @copy_signature(_morpho_method2)
    def inpand(self, clip: vs.VideoNode, *args: Any, func: FuncExceptT | None = None, **kwargs: Any) -> vs.VideoNode:
        return self.xxpand_transform(clip, ExprOp.MIN, *args, func=func, **kwargs)

    @inject_self
    @copy_signature(_morpho_method)
    def closing(self, src: vs.VideoNode, *args: Any, func: FuncExceptT | None = None, **kwargs: Any) -> vs.VideoNode:
        func = func or self.closing

        dilated = self.dilation(src, *args, func=func, **kwargs)
        eroded = self.erosion(dilated, *args, func=func, **kwargs)

        return eroded

    @inject_self
    @copy_signature(_morpho_method)
    def opening(self, src: vs.VideoNode, *args: Any, func: FuncExceptT | None = None, **kwargs: Any) -> vs.VideoNode:
        func = func or self.closing

        eroded = self.erosion(src, *args, func=func, **kwargs)
        dilated = self.dilation(eroded, *args, func=func, **kwargs)

        return dilated

    @inject_self
    def gradient(
        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
        coords: CoordsT = 5, multiply: float | None = None, *, func: FuncExceptT | None = None, **kwargs: Any
    ) -> vs.VideoNode:
        func, planes = self._check_params(radius, thr, coords, planes, func or self.gradient)

        if radius == 1 and self._fast:
            return norm_expr(
                src, '{dilated} {eroded} - {multiply}', planes,
                dilated=self._morpho_xx_imum(src, thr, ExprOp.MAX, coords, None, True),
                eroded=self._morpho_xx_imum(src, thr, ExprOp.MIN, coords, None, True),
                multiply='' if multiply is None else f'{multiply} *'
            )

        eroded = self.erosion(src, radius, planes, thr, coords, multiply, func=func, **kwargs)
        dilated = self.dilation(src, radius, planes, thr, coords, multiply, func=func, **kwargs)

        return norm_expr([dilated, eroded], 'x y -', planes)

    @inject_self
    @copy_signature(_morpho_method)
    def top_hat(self, src: vs.VideoNode, *args: Any, func: FuncExceptT | None = None, **kwargs: Any) -> vs.VideoNode:
        opened = self.opening(src, *args, func=func or self.top_hat, **kwargs)

        return norm_expr([src, opened], 'x y -', kwargs.get('planes', args[1] if len(args) > 1 else None))

    @inject_self
    @copy_signature(_morpho_method)
    def black_hat(self, src: vs.VideoNode, *args: Any, func: FuncExceptT | None = None, **kwargs: Any) -> vs.VideoNode:
        closed = self.closing(src, *args, func=func or self.black_hat, **kwargs)

        return norm_expr([closed, src], 'x y -', kwargs.get('planes', args[1] if len(args) > 1 else None))

    @inject_self
    def outer_hat(
        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
        coords: CoordsT = 5, multiply: float | None = None, *, func: FuncExceptT | None = None, **kwargs: Any
    ) -> vs.VideoNode:
        func, planes = self._check_params(radius, thr, coords, planes, func or self.outer_hat)

        if radius == 1 and self._fast:
            return norm_expr(
                src, '{dilated} {multiply} x -', planes,
                dilated=self._morpho_xx_imum(src, thr, ExprOp.MAX, coords, None, True),
                multiply='' if multiply is None else f'{multiply} *'
            )

        dilated = self.dilation(src, radius, planes, thr, coords, multiply, func=func, **kwargs)

        return norm_expr([dilated, src], 'x y -', planes)

    @inject_self
    def inner_hat(
        self, src: vs.VideoNode, radius: int = 1, planes: PlanesT = None, thr: float | None = None,
        coords: CoordsT = 5, multiply: float | None = None, *, func: FuncExceptT | None = None, **kwargs: Any
    ) -> vs.VideoNode:
        func, planes = self._check_params(radius, thr, coords, planes, func or self.inner_hat)

        if radius == 1 and self._fast:
            return norm_expr(
                src, '{eroded} {multiply} x -', planes,
                eroded=self._morpho_xx_imum(src, thr, ExprOp.MIN, coords),
                multiply='' if multiply is None else f'{multiply} *'
            )

        eroded = self.erosion(src, radius, planes, thr, coords, multiply, func=func, **kwargs)

        return norm_expr([eroded, src], 'x y -', planes)

    @inject_self
    def binarize(
        self, src: vs.VideoNode, midthr: float | list[float] | None = None,
        lowval: float | list[float] | None = None, highval: float | list[float] | None = None,
        planes: PlanesT = None
    ) -> vs.VideoNode:
        midthr, lowval, highval = (
            thr and list(
                scale_value(t, 32, src, chroma=i != 0)
                for i, t in enumerate(to_arr(thr))
            ) for thr in (midthr, lowval, highval)
        )

        return src.std.Binarize(midthr, lowval, highval, planes)


def grow_mask(
    mask: vs.VideoNode, radius: int = 1, multiply: float = 1.0,
    planes: PlanesT = None, coords: CoordsT = 5, thr: float | None = None,
    *, func: FuncExceptT | None = None, **kwargs: Any
) -> vs.VideoNode:
    func = func or grow_mask

    assert check_variable(mask, func)

    morpho = Morpho(planes, func)

    kwargs.update(thr=thr, coords=coords)

    closed = morpho.closing(mask, **kwargs)
    dilated = morpho.dilation(closed, **kwargs)
    outer = morpho.outer_hat(dilated, radius, **kwargs)

    blurred = outer.std.Convolution(wmean_matrix, planes=planes)

    if multiply != 1.0:
        return blurred.std.Expr(f'x {multiply} *')

    return blurred



from ._1d import *  # noqa: F401, F403
from ._2x2 import *  # noqa: F401, F403
from ._3x3 import *  # noqa: F401, F403
from ._5x5 import *  # noqa: F401, F403
from ._abstract import *  # noqa: F401, F403



from __future__ import annotations

from typing import Callable, Protocol, Union

from vstools import CustomEnum, PlanesT, SingleOrArrOpt, vs

from .abstract import GeneralMask
from .edge._abstract import EdgeDetectT, RidgeDetectT

__all__ = [
    'MorphoFunc',
    'XxpandMode',
    'Coordinates',

    'GenericMaskT'
]


class MorphoFunc(Protocol):
    def __call__(
        self, clip: vs.VideoNode, planes: PlanesT = ...,
        threshold: float | None = ..., coordinates: SingleOrArrOpt[int] = ...
    ) -> vs.VideoNode:
        ...


class XxpandMode(CustomEnum):
    """Expand/inpand mode"""

    RECTANGLE = object()
    """Rectangular shape"""

    ELLIPSE = object()
    """Elliptical shape"""

    LOSANGE = object()
    """Diamond shape"""


class Coordinates(list[int], CustomEnum):
    VERTICAL = [0, 1, 0, 0, 0, 0, 1, 0]
    HORIZONTAL = [0, 0, 0, 1, 1, 0, 0, 0]
    RECTANGLE = [1, 1, 1, 1, 1, 1, 1, 1]
    DIAMOND = [0, 1, 0, 1, 1, 0, 1, 0]
    CORNERS = [1, 0, 1, 0, 0, 1, 0, 1]

    @classmethod
    def from_iter(cls, iter: int) -> Coordinates:
        return cls.DIAMOND if (iter % 3) != 1 else cls.RECTANGLE

    @classmethod
    def from_xxpand_mode(cls, xxpand_mode: XxpandMode, iter: int = 1) -> Coordinates:
        if xxpand_mode == XxpandMode.LOSANGE or (xxpand_mode is XxpandMode.ELLIPSE and iter % 3 != 1):
            return cls.DIAMOND

        return cls.RECTANGLE


GenericMaskT = Union[
    vs.VideoNode, Callable[[vs.VideoNode, vs.VideoNode], vs.VideoNode],
    EdgeDetectT, RidgeDetectT, GeneralMask, str
]



from .abstract import *  # noqa: F401, F403
from .details import *  # noqa: F401, F403
from .diff import *  # noqa: F401, F403
from .edge import *  # noqa: F401, F403
from .exceptions import *  # noqa: F401, F403
from .edge_funcs import *  # noqa: F401, F403
from .funcs import *  # noqa: F401, F403
from .hardsub import *  # noqa: F401, F403
from .masks import *  # noqa: F401, F403
from .morpho import *  # noqa: F401, F403
from .types import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Sequence

from vsexprtools import ExprOp, norm_expr
from vsrgtools import box_blur
from vstools import (
    ColorRange, FrameRangeN, FrameRangesN, FramesLengthError, Position, Size, check_variable, depth, get_peak_value,
    normalize_seq, replace_ranges, vs
)

__all__ = [
    'GeneralMask',
    'BoundingBox',
    'DeferredMask'
]


class GeneralMask(ABC):
    @abstractmethod
    def get_mask(self, clip: vs.VideoNode, *args: Any, **kwargs: Any) -> vs.VideoNode:
        ...


class BoundingBox(GeneralMask):
    pos: Position
    size: Size
    invert: bool

    def __init__(self, pos: tuple[int, int] | Position, size: tuple[int, int] | Size, invert: bool = False) -> None:
        self.pos, self.size, self.invert = Position(pos), Size(size), invert

    def get_mask(self, ref: vs.VideoNode, *args: Any, **kwargs: Any) -> vs.VideoNode:  # type: ignore[override]
        from .utils import squaremask
        return squaremask(ref, self.size.x, self.size.y, self.pos.x, self.pos.y, self.invert, self.get_mask)


class DeferredMask(GeneralMask):
    ranges: FrameRangesN
    bound: BoundingBox | None
    refframes: list[int | None]
    blur: bool

    def __init__(
        self, ranges: FrameRangeN | FrameRangesN | None = None, bound: BoundingBox | None = None,
        *, blur: bool = False, refframes: int | list[int | None] | None = None
    ) -> None:
        self.ranges = ranges if isinstance(ranges, Sequence) else [(0, None)] if ranges is None else [ranges]
        self.blur = blur
        self.bound = bound

        if refframes is None:
            self.refframes = []
        else:
            self.refframes = refframes if isinstance(refframes, list) else normalize_seq(refframes, len(self.ranges))

        if len(self.refframes) > 0 and len(self.refframes) != len(self.ranges):
            raise FramesLengthError(
                self.__class__, '', 'Received reference frame and range list size mismatch!'
            )

    def get_mask(self, clip: vs.VideoNode, ref: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:  # type: ignore[override]
        assert check_variable(clip, self.get_mask)
        assert check_variable(ref, self.get_mask)

        if self.bound:
            bm = self.bound.get_mask(ref, **kwargs)

            if self.blur:
                bm = box_blur(bm, 5, 5)

        if len(self.refframes) == 0:
            hm = ref.std.BlankClip(
                format=ref.format.replace(color_family=vs.GRAY, subsampling_h=0, subsampling_w=0).id, keep=True
            )

            for ran, rf in zip(self.ranges, self.refframes):
                if rf is None:
                    rf = ref.num_frames - 1
                elif rf < 0:
                    rf = ref.num_frames - 1 + rf

                mask = depth(
                    self._mask(clip[rf], ref[rf], **kwargs), clip,
                    range_out=ColorRange.FULL, range_in=ColorRange.FULL
                ).std.Loop(hm.num_frames)

                hm = replace_ranges(hm, ExprOp.MAX.combine(hm, mask), ran)
        else:
            hm = depth(
                self._mask(clip, ref, **kwargs), clip,
                range_out=ColorRange.FULL, range_in=ColorRange.FULL
            )

        if self.bound:
            return norm_expr([hm, bm], f'y {ExprOp.clamp(0, get_peak_value(hm), "x")} 0 ?')

        return hm.std.Limiter()

    @abstractmethod
    def _mask(self, clip: vs.VideoNode, ref: vs.VideoNode, **kwargs: Any) -> vs.VideoNode:
        ...
